import prody
import Bio.PDB as bio
import logging
import numpy as np
import re
import sys


# Getting the name of the module for the log system
logger = logging.getLogger(__name__)


def get_ligand_from_PDB(pdb_file):
    """
    :param pdb_file: PDB file with only the ligand
    :return: Bio.PDB object of the input PDB
    """
    parser = bio.PDBParser()
    structure = parser.get_structure("structure", pdb_file)
    return structure


def get_atoms_from_structure(structure):
    """
    :param structure: Bio.PDB object of the input PDB
    :return: list with the atoms that form the input structure
    """
    atom_list = []
    for atom in structure.get_atoms():
        atom_list.append(atom)
    return atom_list


def select_atoms_from_list(PDB_atom_name, atoms_list):
    """
    Given a pdb atom name string and a list of atoms (BioPython Atom) it returns the Bio.Atom correspondent to the atom
    name.
    :param PDB_atom_name: string with an atom name
    :param atoms_list: list of Bio.Atoms
    :return: Bio.Atom correspondent to the atom name
    """
    for atom in atoms_list:
        if atom.name == PDB_atom_name:
            return atom


def get_H_bonded_to_grow(PDB_atom_name, prody_complex, PDB_atom_hydrogen=None, chain="L"):
    """
    Given a heavy atom name (string) and a complex (prody molecule) it returns the hydrogen atom of the chain L
    placed at bonding distance of the input atom name. If there is more than one, a checking of contacts with the
    protein will be performed. In case of finding a possible contact between the hydrogen and the protein, we will
    reject this hydrogen and we will repeat the check in another one. If all the hydrogens have contacts with the
    protein, the first of them will be selected and a warning will be printed.
    :param PDB_atom_name: heavy atom name (string) of a ligand
    :param prody_complex: prody molecule object
    :param PDB_atom_hydrogen: if selected, the name of the specific H atom that you want to bond.
    :return: hydrogen atom of the ligand placed at bonding distance of the heavy atom
    """
    # Select the hydrogens bonded to the heavy atom 'PDB_atom_name'
    selected_h = prody_complex.select("chain {} and hydrogen within 1.70 of name {}".format(chain, PDB_atom_name)) # Replace for 1.53 :)

    # In case that we found more than one we have to select one of them
    try:
        number_of_h = len(selected_h)
        print("Number of hydrogens bonded to {}: {}".format(PDB_atom_name, number_of_h))
    except TypeError:
        raise TypeError("Check either core or fragment atom to bound when passing parameters")
    if len(selected_h) > 1:
        for idx, hydrogen in enumerate(selected_h):
            # We will select atoms of the protein in interaction distance
            select_h_bonds = prody_complex.select("protein and within 2.5 of (name {} and chain L)"
                                                  .format(selected_h.getNames()[idx]))

            if select_h_bonds is not None and PDB_atom_hydrogen is None:
                print("WARNING: {} is forming a close interaction with the protein! We will try to grow"
                               " in another direction.".format(selected_h.getNames()[idx]))
                # We put this elif to select one of H randomly if all of them have contacts
                if (select_h_bonds is not None) and (int(idx) == int(len(selected_h)-1)):
                    hydrogen_pdbatomname = selected_h.getNames()[1]
                    return hydrogen_pdbatomname
            elif select_h_bonds is None and PDB_atom_hydrogen is not None:
                print("Forming a bond between {} and {}...".format(PDB_atom_name, PDB_atom_hydrogen))
                select_specific_h_bonds = selected_h.select("name {}".format(PDB_atom_hydrogen))
                hydrogen_pdbatomname = select_specific_h_bonds.getNames()[0]
                return hydrogen_pdbatomname
            else:
                hydrogen_pdbatomname = selected_h.getNames()[idx]
                return hydrogen_pdbatomname
    else:
        hydrogen_pdbatomname = selected_h.getNames()
        return hydrogen_pdbatomname


# This function is prepared to rename PDB atom names of the repeated names, but is not working currently
def change_repeated_atomnames(list_of_repeated_names, core_names):
    list_of_lists_repeated = []
    for atom in list_of_repeated_names:
        find_name_and_number = re.search('([A-Z]*)([0-9]*)', atom)
        element = find_name_and_number.group(1)
        number = find_name_and_number.group(2)
        list_of_lists_repeated.append([element, number])
    list_of_lists_core = []
    for atom in core_names:
        find_name_and_number = re.search('([A-Z]*)([0-9]*)', atom)
        element = find_name_and_number.group(1)
        number = find_name_and_number.group(2)
        list_of_lists_core.append([element, number])
    for repeated_atom in list_of_lists_repeated:
        for core_atom in list_of_lists_core:
            if repeated_atom[0] is core_atom[0]:
                new_name = [repeated_atom[0], int(repeated_atom[1]) + 1]
                new_name = [new_name[0], str(new_name[1])]
                switch = False
                while not switch:
                    if new_name not in list_of_lists_core:
                        list_of_lists_core.append(new_name)
                        switch = True
                    new_name = [new_name[0], int(new_name[1])+1]
                    new_name = [new_name[0], str(new_name[1])]


def superimpose(fixed_vector, moving_vector, moving_atom_list):
    """
    Rotates and translates a list of moving atoms from a moving vector to a fixed vector.
    :param fixed_vector: vector used as reference.
    :param moving_vector: vector that will rotate and translate.
    :param moving_atom_list: list of atoms that we want to do the rotation and translation of the moving vector.
    :return: the input list of atoms is rotated an translated.
    """
    # Do the superimposition with BioPython
    sup = bio.Superimposer()
    # Set the vectors: first element is the fix vector (bond of the core) and second is the moving (bond of the fragment)
    sup.set_atoms(fixed_vector, moving_vector)
    # Apply the transformation to the atoms of the fragment (translate and rotate)
    return sup.apply(moving_atom_list)


def transform_coords(atoms_with_coords):
    """
    Transform the coords of a molecule (ProDy selection) into the coords from a list of atoms of Bio.PDB.
    :param atoms_with_coords: list of atoms (from a Bio.PDB) with the coordinates that we want to set.
    :return: perform the transformation of the coords.
    """
    coords = []
    for atom in atoms_with_coords:
        coords.append(list(atom.get_coord()))
    return np.asarray(coords)


def extract_and_change_atomnames(molecule, selected_resname, core_resname, rename=False):
    """
    Given a ProDy molecule and a Resname this function will rename the PDB atom names for the selected residue following
    the next pattern: G1, G2, G3...
    :param molecule: ProDy molecule.
    :param selected_resname: Residue name whose atoms you would like to rename.
    :return: ProDy molecule with atoms renamed and dictionary {"original atom name" : "new atom name"}
    """
    assert selected_resname != core_resname, "core and fragment residue name must be different"
    fragment = molecule.select("resname {}".format(selected_resname))
    core = molecule.select("resname {}".format(core_resname))
    core_atom_names  = [atom.getName() for atom in core]
    fragment_atom_names  = [atom.getName() for atom in fragment]
    names_dictionary = {}
    for n, atom_name in enumerate(fragment_atom_names):
        if rename:
            names_dictionary[atom_name] = "G{}".format(n)
        else:
            # If the atomname is repited
            if atom_name in core_atom_names:
                initial_atom_name = atom_name
                while atom_name in core_atom_names:
                    atom_name_digit = re.findall('\d+', atom_name)[0]
                    new_atom_name_digit = int(atom_name_digit) + 1
                    atom_name = atom_name.replace(atom_name_digit, str(new_atom_name_digit))
                final_atom_name = atom_name
                core_atom_names.append(final_atom_name)
                names_dictionary[initial_atom_name] = final_atom_name
            else:
                names_dictionary[atom_name] = atom_name
    for atom in molecule:
        if atom.getResname() == selected_resname:
            if atom.getName() in names_dictionary:
                atom.setName(names_dictionary[atom.getName()])
    return molecule, names_dictionary


def check_overlapping_names(structures_to_bond):
    """
    Checking that there is not duplications in the names of the structure. If not, it will return None, else, it will
    return the repeated elements.
    :param structures_to_bond: ProDy molecule
    :return: set object with the repeated elements if they are found. Else, None object.
    """
    all_atom_names = list(structures_to_bond.getNames())
    return set([name for name in all_atom_names if all_atom_names.count(name) > 1])








# General imports
import sys
import os
import time
import subprocess
import glob
import shutil
import logging
# Local imports
import frag_pele.constants as c
import frag_pele.Helpers.folder_handler
import frag_pele.Helpers.clusterizer
import frag_pele.Growing.add_fragment_from_pdbs
import frag_pele.Growing.simulations_linker
import frag_pele.Growing.bestStructs


# Getting the name of the module for the log system
logger = logging.getLogger(__name__)
curr_dir = os.path.abspath(os.path.curdir)
FilePath = os.path.relpath(__file__)


def run_no_grow(complex_pdb, sch_python, resfold, iterations, cpus, ID, pele_dir,
                pdbout, restart, min_overlap, max_overlap, contrl, criteria, report,
                epsilon, clusterThreshold, distance_contact, condition, traject,
                metricweights, nclusters, pele_eq_steps, license):
    # Time computations
    start_time = time.time()
    pdbout_folder = "{}_{}".format(pdbout, ID)

    # Path definition
    plop_relative_path = os.path.join("/".join(FilePath.split("/")[:-2]), c.PLOP_PATH)
    Helpers.folder_handler.check_and_create_DataLocal()

    # Get ligand and extract templates
    pdb_input = os.path.join(c.PRE_WORKING_DIR, complex_pdb)
    ligand_name = Growing.add_fragment_from_pdbs.extract_heteroatoms_pdbs(pdb_input)
    pdb_to_template = "{}.pdb".format(ligand_name)
    shutil.copy(pdb_to_template, c.TEMPLATES_PATH)
    cmd = "{} {} {}".format(sch_python, plop_relative_path, os.path.join(curr_dir, c.TEMPLATES_PATH, pdb_to_template))
    subprocess.call(cmd.split())
    template_name = "{}z".format(ligand_name.lower())
    shutil.copy(template_name, os.path.join(curr_dir, c.TEMPLATES_PATH))
    shutil.copy("{}.rot.assign".format(ligand_name), os.path.join(curr_dir, c.ROTAMERS_PATH))

    # Preparation of results and output folders
    results = ["{}{}_{}{}".format(c.OUTPUT_FOLDER, ID, resfold, n) for n in range(0, iterations + 1)]
    pdbs = [complex_pdb if n == 0 else "{}_{}".format(n, complex_pdb) for n in range(0, iterations + 1)]
    pdb_selected_names = ["initial_0_{}.pdb".format(n) for n in range(0, int(cpus) - 1)]

    if not restart:
        list_of_subfolders = glob.glob("{}*".format(pdbout_folder))
        for subfolder in list_of_subfolders:
            shutil.rmtree(subfolder)
    # Simulation loop - LOOP CORE
    for i, (pdb_file, result) in enumerate(zip(pdbs, results)):
        # Only if reset
        if restart:
            if os.path.exists(os.path.join(pdbout_folder, "{}".format(i))) and os.path.exists(os.path.join(pdbout_folder,
                                                                                                    "{}".format(i),
                                                                                                    "initial_0_0.pdb")):
                print("STEP {} ALREADY DONE, JUMPING TO THE NEXT STEP...".format(i))
                continue
        # Otherwise start from the beggining
        pdb_input_paths = ["{}".format(os.path.join(pdbout_folder, str(i - 1), pdb_file)) for pdb_file in
                           pdb_selected_names]

        # Control file modification
        overlapping_factor = float(min_overlap) + (((float(max_overlap) - float(min_overlap))*i) / iterations)
        overlapping_factor = "{0:.2f}".format(overlapping_factor)

        if i != 0:
            logger.info(c.SELECTED_MESSAGE.format(contrl, pdb_selected_names, result, i))
            Growing.simulations_linker.control_file_modifier(contrl, pdb_input_paths, i, license, overlapping_factor,
                                                             result)
        else:
            logger.info(c.SELECTED_MESSAGE.format(contrl, os.path.join(c.PRE_WORKING_DIR, complex_pdb), result, i))
            Growing.simulations_linker.control_file_modifier(contrl, [os.path.join(c.PRE_WORKING_DIR, complex_pdb)],
                                                             i, license, overlapping_factor,
                                                             result) #  We have put [] in pdb_initialize
                                                                     #  because by default we have to use
                                                                     #  a list as input
        logger.info(c.LINES_MESSAGE)

        # Running PELE simulation
        Helpers.folder_handler.check_and_create_results_folder(result)
        Growing.simulations_linker.simulation_runner(pele_dir, contrl, int(cpus))
        logger.info(c.LINES_MESSAGE)
        logger.info(c.FINISH_SIM_MESSAGE.format(result))
        # Before selecting a step from a trajectory we will save the input PDB file in a folder
        Helpers.folder_handler.check_and_create_pdb_clusters_folder(pdbout_folder, i)
        # ---------------------------------------------------CLUSTERING-------------------------------------------------
        # Transform column name of the criteria to column number
        result_abs = os.path.abspath(result)
        logger.info("Looking structures to cluster in '{}'".format(result_abs))
        column_number = Helpers.clusterizer.get_column_num(result_abs, criteria, report)
        # Selection of the trajectory used as new input

        Helpers.clusterizer.cluster_traject(str(ligand_name), int(cpus - 1), column_number, distance_contact,
                                            clusterThreshold, "{}*".format(os.path.join(result_abs, traject)),
                                            os.path.join(pdbout_folder, str(i)), epsilon, report, condition,
                                            metricweights, nclusters)

    # ----------------------------------------------------EQUILIBRATION-------------------------------------------------
    # Set input PDBs
    pdb_inputs = ["{}".format(os.path.join(pdbout_folder, str(iterations), pdb_file)) for pdb_file in pdb_selected_names]
    logger.info(".....STARTING EQUILIBRATION.....")
    if not os.path.exists("equilibration_result_{}".format(ID)):  # Create the folder if it does not exist
        os.mkdir("equilibration_result_{}".format(ID))
    # Modify the control file to increase the steps to 20 and change the output path
    Growing.simulations_linker.control_file_modifier(contrl, pdb_inputs, iterations, license, max_overlap,
                                                     "equilibration_result_{}".format(ID), pele_eq_steps)
    # Call PELE to run the simulation
    Growing.simulations_linker.simulation_runner(pele_dir, contrl, int(cpus))
    equilibration_path = os.path.join(os.path.abspath(os.path.curdir), "equilibration_result_{}".format(ID))
    if not os.path.exists("selected_result_{}".format(ID)):  # Create the folder if it does not exist
        os.mkdir("selected_result_{}".format(ID))
    os.chdir("selected_result_{}".format(ID))
    Growing.bestStructs.main(criteria, "best_structure.pdb", path=equilibration_path, n_structs=10)
    shutil.copy("sel_0_best_structure.pdb", "../pregrow/selection_{}.pdb".format(ID))
    os.chdir("../")
    end_time = time.time()
    total_time = (end_time - start_time) / 60
    logging.info("Simulation of {} performed in {} min".format(complex_pdb, total_time))



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import range
import numpy as np
import matplotlib.pyplot as plt
from scipy import sparse, linalg
from AdaptivePELE.freeEnergies import utils


def buildLinearChainCountMatrix():
    C = np.array([[6000., 3, 0, 0, 0, 0],
                  [3, 1000., 3, 0, 0, 0],
                  [0, 3, 1000., 3, 0, 0],
                  [0, 0, 3, 1000., 3, 0],
                  [0, 0, 0, 3, 1000., 3],
                  [0, 0, 0, 0, 3, 90000.]])
    return C+1/6.


def buildPCountMatrix():
    C = np.array([[600., 2, 2, 0, 0, 0],
                  [2, 100., 0, 2, 2, 0],
                  [2, 0, 100., 2, 2, 0],
                  [0, 2, 2, 100., 0, 2],
                  [0, 2, 2, 0, 100., 2],
                  [0, 0, 0, 2, 2, 9000.]])
    return C+1/6.


def normaliseMatrixVectors(M):
    normArray = []
    for x in M:
        normArray.append(x / np.linalg.norm(x))
    return np.array(normArray)


def normaliseMatrixRows(M):
    sums = M.sum(axis=1)
    T = M / sums[:, np.newaxis]
    return T


def buildTransitionMatrix(C):
    return normaliseMatrixRows(C)


def buildRevTransitionMatrix_fast(C, iterations=5):
    """
        #Implemented as Prinz paper
    """
    X = C + C.T
    x = X.sum(axis=1)
    c = C.sum(axis=1)

    return loop(iterations, X, x, C, c)


def loop(iterations, X, x, C, c):

    n = len(x)

    c_vec = c[:, np.newaxis]
    A = c_vec - C + c_vec.T - C.T

    for _ in range(iterations):
        for i in range(n):
            X[i, i] = C[i, i] * (x[i] - X[i, i]) / (c[i] - C[i, i])

        x = X.sum(axis=1)

        x_vec = x[:, np.newaxis]

        B = c_vec*(x_vec.T - X) + c_vec.T*(x_vec - X) - (C + C.T) * (x_vec + x_vec.T - 2 * X)
        Z = -(C+C.T) * (x_vec - X) * (x_vec.T - X)

        indicesU = np.triu_indices(n, 1)
        X[indicesU] = (-B[indicesU] + np.sqrt(B[indicesU]**2 - 4 * A[indicesU]*Z[indicesU])) / (2 * A[indicesU])
        X.T[indicesU] = X[indicesU]

        x = X.sum(axis=1)

    T = np.zeros((n, n))
    for i in range(n):
        for j in range(n):
            T[i, j] = X[i, j] / x[i]
    return T
    # return X / x[:,np.newaxis]


def buildRevTransitionMatrix(C):
    """
        #Implemented as Prinz paper
    """
    X = C + C.T
    x = X.sum(axis=1)
    c = C.sum(axis=1)

    iterations = 1000
    n = len(x)

    T = np.zeros(C.shape)
    for _ in range(iterations):
        # if it != 0 and it % 10 == 0:
        #    print(it)
        for i in range(n):
            X[i, i] = C[i, i] * (x[i] - X[i, i]) / (c[i] - C[i, i])
        x = X.sum(axis=1)
        for i in range(n - 1):
            for j in range(i + 1, n):
                a = c[i] - C[i, j] + c[j] - C[j, i]
                b = c[i] * (x[j]-X[i, j]) + c[j] * (x[i] - X[i, j])-(C[i, j] + C[j, i])*(x[i] + x[j] - 2*X[i, j])
                z = -(C[i, j] + C[j, i])*(x[i] - X[i, j])*(x[j] - X[i, j])
                X[i, j] = X[j, i] = (-b + np.sqrt(b**2 - 4*a*z))/(2*a)
        x = X.sum(axis=1)

    for i in range(n):
        for j in range(n):
            T[i, j] = X[i, j] / x[i]
    return T


def runSimulation(P, steps, startingPosition):
    n = P.shape[0]
    position = startingPosition

    traj = np.zeros(steps)
    traj[0] = position
    for step in range(steps):
        prob = P[position]
        position = np.random.choice(range(n), p=prob)
        traj[step] = position

    return traj


def runSetOfSimulations(numberOfSimulations, P, steps):
    trajs = []
    for sim in range(numberOfSimulations):
        if sim % 50 == 0:
            print("Simulation", sim)

        startingPosition = sim % P.shape[0]
        traj = runSimulation(P, steps, startingPosition)
        trajs.append(traj)
    return trajs

# def estimateCountMatrix(trajectories, n, tau):
#     counts = np.zeros((n, n))
#
#     #for traj in trajectories:
#     for i, traj in enumerate(trajectories):
#         print(i)
#         for i in range(len(traj) - tau):
#             fromState = traj[i]
#             toState = traj[i + tau]
#
#             counts[fromState][toState] += 1
#     return counts


def estimateCountMatrix(trajectories, n, tau):
    rows = []
    cols = []
    for traj in trajectories:
        if traj.size > tau:
            rows.append(traj[0:-tau])
            cols.append(traj[tau:])

    if len(rows) == 0:
        raise ValueError('Too long lagtime')

    row = np.concatenate(rows)
    col = np.concatenate(cols)
    data = np.ones(row.size)
    C = sparse.coo_matrix((data, (row, col)), shape=(n, n))
    return C.toarray()


def printMatrix(matrix):
    for array in matrix:
        for element in array:
            print("%e" % element)
        print("")


def estimateTransitionMatrix(trajectories, n, tau, symm=True):
    C = estimateCountMatrix(trajectories, n, tau) + 1./n
    if symm:
        return utils.buildRevTransitionMatrix(C)
    return buildTransitionMatrix(C)


def getSortedEigen(T):
    eigenvals, eigenvectors = linalg.eig(T, left=True, right=False)
    sortedIndices = np.argsort(eigenvals)[::-1]

    # reigenvals, reigenvectors = linalg.eig(T.T)
    # rsortedIndices = np.argsort(reigenvals)[::-1]
    return eigenvals[sortedIndices], eigenvectors[:, sortedIndices]


def getSortedEigenFromDtrajs(tau, trajs, n):
    estimatedT = estimateTransitionMatrix(trajs, n, tau)
    # T.T*pi = pi; or pi*T = pi, where pi is col and row array respectively
    return getSortedEigen(estimatedT)


def getStationaryDistr(lowestEigenvector):
    absStationary = np.abs(lowestEigenvector)
    return absStationary / absStationary.sum()


def getStationaryDistrFromTransition(transitionMatrix):
    _, eic = getSortedEigen(transitionMatrix)
    return getStationaryDistr(eic[:, 0])


def getRelativeEntropy(goldenStationary, goldenT, T):
    return np.dot(goldenStationary, goldenT*np.log(goldenT/T)).sum()


def getGoldenTForGivenTau(T, tau):
    n = T.shape[0]
    accT = np.eye(n)
    for _ in range(tau):
        accT = np.dot(accT, T)
    return accT


def plotEigenvalEvolutionInTau(trajs, taus, n):
    allEigenvals = []
    for i, tau in enumerate(taus):
        eigenvals, _ = getSortedEigenFromDtrajs(tau, trajs, n)
        allEigenvals.append(eigenvals)

    allEigenvals = np.array(allEigenvals)  # rework

    fig = plt.figure(1)
    ax = fig.add_subplot(1, 1, 1)
    ax.set_yscale('log')
    valuesToPlot = n-1
    for i in range(1, valuesToPlot):
        ax.plot(taus, allEigenvals[:, i])
    plt.figure(2)
    for i in range(1, valuesToPlot):
        plt.plot(taus, -taus/np.log(allEigenvals[:, i]))


def main():
    numberOfStates = 6
    # steps = 50
    # numberOfSimulations = 60000
    # taus = np.array([1,10,20,50,100,200,500])
    # taus = np.array([1])

    n = numberOfStates

    C = buildLinearChainCountMatrix()
    T = buildTransitionMatrix(C)

    eigenvals, eigenvectors = linalg.eig(T, left=True, right=False)
    sortedIndices = np.argsort(eigenvals)[::-1]
    # stationary distribution
    goldenStationary = getStationaryDistr(eigenvectors[:, sortedIndices[0]])

    print("real C")
    print(printMatrix(C))
    print("real T")
    print(printMatrix(T))
    print("=====")

    # trajs = runSetOfSimulations(numberOfSimulations, T, steps)
    # np.save("trajs_S_50_60000.npy", trajs)
    # np.save("trajs_chain_small.npy", trajs)
    trajs = np.load("trajs_S_50_60000.npy")

    taus = np.array(range(1, 49))
    # taus = np.array([1, 10, 25,  50,  75,  100,  250,  500,  750,  1000,  2000,  2500])
    plotEigenvalEvolutionInTau(trajs, taus, n)

    simLengths = list(range(50, 2000, 50))
    simLengths = list(range(10, 50, 5))
    trajNumbers = list(range(0, 600, 5))
    trajNumbers = [600]

    entropies = np.zeros(len(simLengths))
    tau = 1
    for trajNum in trajNumbers:
        for i, j in enumerate(simLengths):
            trimmedTrajs = trajs[0:trajNum, :j]
            eigenvals, _ = getSortedEigenFromDtrajs(tau, trimmedTrajs, n)
            # stationary = getStationaryDistr(eigenvec[:,0])

            estimatedT = estimateTransitionMatrix(trimmedTrajs, n, tau)
            goldenT = getGoldenTForGivenTau(T, tau)

            rEntropy = getRelativeEntropy(goldenStationary, goldenT, estimatedT)
            print(j, rEntropy)
            entropies[i] = rEntropy

    fig = plt.figure(3)
    ax = fig.add_subplot(1, 1, 1)
    ax.set_yscale('log')
    plt.plot(simLengths, entropies)
    plt.show()

if __name__ == "__main__":
    main()



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import range
import math
import sys
import numpy as np
import random
import scipy.optimize as optim
import os
import glob
from AdaptivePELE.constants import blockNames
from AdaptivePELE.constants import constants
from AdaptivePELE.utilities import utilities
from AdaptivePELE.spawning import spawningTypes
from AdaptivePELE.spawning import densitycalculator
from abc import abstractmethod


def reward(x, rews):
    return -(x[:, np.newaxis]*rews).sum()


def return_sign(i, m, n, r):
    """
        Helper function, creates a three-piece step function

        :param i: Value to compare
        :type i: int
        :param m: Middle value
        :type m: int
        :param n: Left extreme value
        :type n: int
        :param r: Right extreme value
        :type r: int

        :returns: int -- Three-piece sign
    """

    if i <= n-m:
        return 1
    elif i <= r:
        return 0
    else:
        return -1


def getSizes(clusters):
    """
        Get the size of the clusters

        :param clusters: Existing clusters
        :type clusters: :py:class:`.Clusters`

        :returns: np.Array -- Array containing the size of the clusters
    """
    sizes = np.zeros(len(clusters))
    for i, cluster in enumerate(clusters):
        sizes[i] = cluster.elements
    return sizes


def calculateContactsVar(deltaR, epsMax):
    """
        Calculate the variation of epsilon according to the contact ratio

        :param deltaR: Change in contact ratio
        :type deltaR: float
        :param epsMax: Maximum value of epsilon
        :type epsMax: float

        :returns: float -- Epsilon variation
    """
    if deltaR < 0.1:
        return 0
    elif deltaR > 1.0:
        return epsMax * 0.09
    else:
        return epsMax * 0.09 * deltaR


class SpawningAlgorithmBuilder:

    def build(self, spawningBlock):
        """
            Build the selected spawning calculator and spawning params objects

            :param spawningBlock: Block of the control file with the spawning
                parameters
            :type spawningBlock: dict

            :returns: :py:class:`.SpawningCalculator`,
                :py:class:`.SpawningParams` -- SpawningCalculator and
                SpawningParams objects
        """
        spawningCalculatorBuilder = SpawningBuilder()
        spawningCalculator = spawningCalculatorBuilder.buildSpawningCalculator(spawningBlock)

        spawningParams = SpawningParams()
        spawningParams.buildSpawningParameters(spawningBlock)

        return spawningCalculator, spawningParams


class SpawningBuilder:

    def buildSpawningCalculator(self, spawningBlock):
        """
            Build the selected spawning calculator object

            :param spawningBlock: Block of the control file with the spawning
                parameters
            :type spawningBlock: dict

            :returns: :py:class:`.SpawningCalculator` -- SpawningCalculator
                object
        """
        densityBuilder = densitycalculator.DensityCalculatorBuilder()
        densityCalculator = densityBuilder.build(spawningBlock)

        spawningTypeString = spawningBlock[blockNames.StringSpawningTypes.type]
        if spawningTypeString == blockNames.StringSpawningTypes.sameWeight:
            spawningCalculator = SameWeightDegeneracyCalculator()
        elif spawningTypeString == blockNames.StringSpawningTypes.independent:
            spawningCalculator = IndependentRunsCalculator()
        elif spawningTypeString == blockNames.StringSpawningTypes.inverselyProportional:
            spawningCalculator = InverselyProportionalToPopulationCalculator(densityCalculator)
        elif spawningTypeString == blockNames.StringSpawningTypes.epsilon:
            spawningCalculator = EpsilonDegeneracyCalculator(densityCalculator)
        elif spawningTypeString == blockNames.StringSpawningTypes.fast:
            spawningCalculator = FASTDegeneracyCalculator(densityCalculator)
        elif spawningTypeString == blockNames.StringSpawningTypes.variableEpsilon:
            spawningCalculator = VariableEpsilonDegeneracyCalculator(densityCalculator)
        elif spawningTypeString == blockNames.StringSpawningTypes.UCB:
            spawningCalculator = UCBCalculator(densityCalculator)
        elif spawningTypeString == blockNames.StringSpawningTypes.REAP:
            spawningCalculator = REAPCalculator()
        elif spawningTypeString == blockNames.StringSpawningTypes.null:
            spawningCalculator = NullSpawningCalculator()
        else:
            sys.exit("Unknown spawning type! Choices are: " + str(spawningTypes.SPAWNING_TYPE_TO_STRING_DICTIONARY.values()))
        return spawningCalculator


class SpawningParams:

    def __init__(self):
        self.epsilon = None
        self.temperature = None
        self.threshold = None
        self.reportFilename = None
        self.reportCol = None
        self.decrement = None
        self.varEpsilonType = None
        self.maxEpsilon = None
        self.minEpsilon = None
        self.variationWindow = None
        self.maxEpsilonWindow = None
        self.metricWeights = None
        self.alpha = None
        self.nclusters = None  # number of clusters to consider in epsilon
        self.period = None
        self.metricInd = None
        self.condition = blockNames.SpawningParams.minValue  # wether to consider min or max values in epsilon

    def buildSpawningParameters(self, spawningBlock):
        """
            Build the selected spawning params objects

            :param spawningBlock: Block of the control file with the spawning
                parameters
            :type spawningBlock: dict

            :returns: :py:class:`.SpawningParams` -- SpawningParams object
        """
        spawningParamsBlock = spawningBlock[blockNames.SpawningParams.params]
        spawningType = spawningBlock[blockNames.StringSpawningTypes.type]

        # Params general to all
        # reportFilename is now mandatory for all spawning
        self.reportFilename = spawningParamsBlock[blockNames.SpawningParams.report_filename]
        # Params specific to epsilon related spawning
        if spawningType == blockNames.StringSpawningTypes.epsilon or \
                spawningType == blockNames.StringSpawningTypes.variableEpsilon:
            self.epsilon = spawningParamsBlock[blockNames.SpawningParams.epsilon]
            self.nclusters = spawningParamsBlock.get(blockNames.SpawningParams.nclusters, 5)
            self.metricWeights = spawningParamsBlock.get(blockNames.SpawningParams.metricWeights,
                                                         blockNames.SpawningParams.linear)
            self.condition = spawningParamsBlock.get(blockNames.SpawningParams.condition,
                                                     blockNames.SpawningParams.minValue)

        if spawningType == blockNames.StringSpawningTypes.epsilon or \
                spawningType == blockNames.StringSpawningTypes.variableEpsilon or\
                spawningType == blockNames.StringSpawningTypes.fast or \
                spawningType == blockNames.StringSpawningTypes.simulatedAnnealing or \
                spawningType == blockNames.StringSpawningTypes.UCB or \
                spawningType == blockNames.StringSpawningTypes.REAP:
            self.temperature = spawningParamsBlock.get(blockNames.SpawningParams.temperature, 1000)
            # Start counting the columns by 1
            self.reportCol = spawningParamsBlock[blockNames.SpawningParams.report_col]-1

        if spawningType == blockNames.StringSpawningTypes.variableEpsilon:
            self.varEpsilonType = spawningParamsBlock[blockNames.SpawningParams.varEpsilonType]
            self.maxEpsilon = spawningParamsBlock[blockNames.SpawningParams.maxEpsilon]
            if self.varEpsilonType == blockNames.VariableEpsilonTypes.linearVariation:
                self.minEpsilon = spawningParamsBlock.get(blockNames.SpawningParams.minEpsilon, self.epsilon)
                self.variationWindow = spawningParamsBlock[blockNames.SpawningParams.variationWindow]
                self.maxEpsilonWindow = spawningParamsBlock[blockNames.SpawningParams.maxEpsilonWindow]
                self.period = spawningParamsBlock.get(blockNames.SpawningParams.period, self.variationWindow)
                self.period += np.sign(np.abs(self.variationWindow-self.period))
                # Add one epoch to the total lenght of the variation in the case of periodic variation to leave a step between variation periods

        if spawningType == blockNames.StringSpawningTypes.UCB:
            self.alpha = spawningParamsBlock.get(blockNames.SpawningParams.alpha, 8.0)

        if spawningType == blockNames.StringSpawningTypes.REAP:
            self.metricInd = spawningParamsBlock.get(blockNames.SpawningParams.metricsInd, -1)


class SpawningCalculator:
    """
        The purpose of this abstract class is to contain the behaviour of the different strategies for the spawning.
        Spawning is the way in which we split the different explorers at the begining of each epoch.
    """

    def __init__(self):
        self.type = "BaseClass"  # change for abstract attribute

    @abstractmethod
    def calculate(self, clusters, trajToDivide, spawningParams, currentEpoch=None):
        pass

    @abstractmethod
    def log(self):
        """
            Log spawning information
        """
        pass

    def writeSpawningInitialStructures(self, outputPathConstants, degeneracyOfRepresentatives, clustering, iteration):
        """
            Write initial structures for the next iteration

            :param outputPathConstants: Output constants that depend on the path
            :type outputPathConstants: :py:class:`.OutputPathConstants`
            :param degeneracyOfRepresentatives: List with the degeneracy of
                each cluster (number of processors that will start from that state)
            :type degeneracyOfRepresentatives: list
            :param clustering: Clustering object
            :type clustering: :py:class:`.Clustering`
            :param iteration: Number of epoch
            :type iteration: int

            :returns: int, list -- number of processors, list with the
                snapshot from which the trajectories will start in the next iteration
        """
        tmpInitialStructuresTemplate = outputPathConstants.tmpInitialStructuresTemplate
        counts = 0
        procMapping = []
        for i, cluster in enumerate(clustering.clusters.clusters):
            for _ in range(int(degeneracyOfRepresentatives[i])):
                outputFilename = tmpInitialStructuresTemplate % (iteration, counts)
                print('Writing to ', outputFilename, 'cluster', i)
                procMapping.append(cluster.writeSpawningStructure(outputFilename))

                counts += 1

        print("counts & cluster centers", counts, np.where(np.array(degeneracyOfRepresentatives) > 0)[0].size)
        return counts, procMapping

    def divideTrajAccordingToWeights(self, weights, trajToDistribute):
        """
            Distribute the trajectories among the clusters according to their
            weight. Weights must be normalized (i.e. sum(weights) = 1)

            :param weights: Weight of each cluster
            :type weights: np.Array
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int

            :returns: list -- List with the number of processors allocated to
                each cluster
        """
        degeneracy = []
        for i, weight in enumerate(weights):
            degeneracy.append(int(weight*trajToDistribute))

        # divide remaining traj to distribute according to decimal part
        decimalPart = []
        decimalPart = [math.modf(weight*trajToDistribute)[0] for weight in weights]
        sortedDecimals = np.argsort(decimalPart)
        sortedDecimals = sortedDecimals[::-1]  # flip list

        leftProcessors = trajToDistribute-sum(degeneracy)
        for i in range(leftProcessors):
            degeneracy[sortedDecimals[i]] += 1

        return degeneracy

    def divideProportionalToArray(self, array, trajToDistribute):
        """
            Distribute the trajectories among the clusters according to the
            values of an array.

            :param Array: Weight of each cluster
            :type Array: np.Array
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int

            :returns: list -- List with the number of processors allocated to
                each cluster
        """
        if isinstance(array, list):
            array = np.array(array)
        weights = array/sum(array)
        return self.divideTrajAccordingToWeights(weights, trajToDistribute)

    def divideInverselyProportionalToArray(self, array, trajToDistribute):
        """
            Distribute the trajectories among the clusters inversely proportional
            to the values of an array.

            :param Array: Weight of each cluster
            :type Array: np.Array
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int

            :returns: list -- List with the number of processors allocated to
                each cluster
        """
        if isinstance(array, list):
            array = np.array(array)
        weights = 1./array

        # Handle Nan cases
        weights[weights == np.inf] = 0

        # Handle all Nan cases
        if weights.any():
            weights /= sum(weights)
        else:
            weights[:] = 1./weights.shape[0]

        return self.divideTrajAccordingToWeights(weights, trajToDistribute)

    def getMetrics(self, clusters):
        """
            Get the metric of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`

            :returns: np.Array -- Array containing the metric of the clusters
        """
        metrics = np.zeros(len(clusters))
        for i, cluster in enumerate(clusters):
            metrics[i] = cluster.getMetric()
        return metrics


class DensitySpawningCalculator(SpawningCalculator):
    """
        Subclass of Spawning calculator that ensures the definition of a density calculator.
    """

    def __init__(self, densityCalculator=densitycalculator.NullDensityCalculator()):
        SpawningCalculator.__init__(self)
        self.type = "BaseDensityClass"  # change for abstract attribute
        self.densityCalculator = densityCalculator

    def calculateDensities(self, clusters):
        """
            Calculate the densities of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`

            :returns: np.Array -- Array containing the density of the clusters
        """
        densities = np.zeros(len(clusters))
        for i, cluster in enumerate(clusters):
            contacts = cluster.getContacts()
            cluster.density = self.densityCalculator.calculate(contacts, cluster.contactThreshold)
            densities[i] = cluster.density
        return densities


class IndependentRunsCalculator(SpawningCalculator):

    def __init__(self):
        SpawningCalculator.__init__(self)
        self.calculator = SameWeightDegeneracyCalculator()
        self.type = spawningTypes.SPAWNING_TYPES.independent

    def calculate(self, clusters, trajToDistribute, spawningParams,
                  currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        return self.calculator.calculate(clusters, trajToDistribute, spawningParams, currentEpoch)

    def log(self):
        """
            Log spawning information
        """
        pass

    def writeSpawningInitialStructures(self, outputPathConstants, degeneracyOfRepresentatives, clustering, iteration):
        """
            Write last trajectory structure as initial one for the next iteration

            :param outputPathConstants: Output constants that depend on the path
            :type outputPathConstants: :py:class:`.OutputPathConstants`
            :param degeneracyOfRepresentatives: List with the degeneracy of
                each cluster (number of processors that will start from that state)
            :type degeneracyOfRepresentatives: list
            :param clustering: Clustering object
            :type clustering: :py:class:`.Clustering`
            :param iteration: Number of epoch
            :type iteration: int

            :returns: int, list -- number of processors, list with the
                snapshot from which the trajectories will start in the next iteration
        """
        procMapping = []
        trajWildcard = os.path.join(outputPathConstants.epochOutputPathTempletized, constants.trajectoryBasename)
        trajectories = glob.glob(trajWildcard % (iteration-1))
        for num, trajectory in enumerate(trajectories):
            snapshots = utilities.getSnapshots(trajectory)
            lastSnapshot = snapshots[-1]
            nSnapshots = len(snapshots)
            del snapshots

            numTraj = int(os.path.splitext(trajectory.rsplit("_", 1)[-1])[0])
            outputFilename = outputPathConstants.tmpInitialStructuresTemplate % (iteration, num)
            procMapping.append((iteration-1, numTraj, nSnapshots-1))

            with open(outputFilename, 'w') as f:
                f.write(lastSnapshot)

        return len(trajectories), procMapping


class SameWeightDegeneracyCalculator(SpawningCalculator):

    def __init__(self):
        SpawningCalculator.__init__(self)
        self.type = spawningTypes.SPAWNING_TYPES.sameWeight

    def calculate(self, clusters, trajToDistribute, spawningParams,
                  currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        numClusters = len(clusters)
        trajToDistribute = min(trajToDistribute, numClusters)
        samples = random.sample(range(numClusters), trajToDistribute)
        degeneracy = [0] * len(clusters)
        for sample in samples:
            degeneracy[sample] = 1
        return degeneracy

    def log(self):
        """
            Log spawning information
        """
        pass


class InverselyProportionalToPopulationCalculator(DensitySpawningCalculator):

    def __init__(self, densityCalculator=densitycalculator.NullDensityCalculator()):
        DensitySpawningCalculator.__init__(self, densityCalculator)
        self.type = spawningTypes.SPAWNING_TYPES.inverselyProportional

    def log(self):
        """
            Log spawning information
        """
        pass

    def calculate(self, clusters, trajToDistribute, spawningParams, currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        sizes = getSizes(clusters)
        densities = self.calculateDensities(clusters)

        if densities.any():
            weights = sizes/densities
        else:
            weights = sizes

        argweights = weights.argsort()
        weights_trimmed = np.zeros(len(sizes)) + 1e6
        weights_trimmed[argweights[:trajToDistribute]] = weights[argweights[:trajToDistribute]]
        return self.divideInverselyProportionalToArray(weights_trimmed, trajToDistribute)


class EpsilonDegeneracyCalculator(DensitySpawningCalculator):
    """
        It uses epsilon * numTraj trajectories proportional to their energy and the rest inversely proportional to each cluster's population
        We only consider the nclusters with best metric
    """

    def __init__(self, densityCalculator=densitycalculator.NullDensityCalculator(), condition=blockNames.SpawningParams.minValue):
        DensitySpawningCalculator.__init__(self, densityCalculator)
        self.inverselyProportionalCalculator = InverselyProportionalToPopulationCalculator(densityCalculator)
        self.type = spawningTypes.SPAWNING_TYPES.epsilon
        self.degeneracyInverselyProportional = None
        self.degeneracyMetricProportional = None
        self.degeneracyTotal = None
        self.condition = condition

    def log(self):
        """
            Log spawning information
        """
        if self.degeneracyTotal is not None:
            print("[SpawningLog] Total: %s" % str(self.degeneracyTotal))
        if self.degeneracyInverselyProportional is not None:
            print("[SpawningLog] Inversely prop: %s" % str(self.degeneracyInverselyProportional))
        if self.degeneracyMetricProportional is not None:
            print("[SpawningLog] Metric prop:    %s" % str(self.degeneracyMetricProportional))

    def calculate(self, clusters, trajToDistribute, spawningParams, currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        trajToMetricProportional = int(spawningParams.epsilon * trajToDistribute)
        trajToInverselyProportional = trajToDistribute - trajToMetricProportional

        self.degeneracyInverselyProportional = self.inverselyProportionalCalculator.calculate(clusters, trajToInverselyProportional, spawningParams)
        self.degeneracyMetricProportional = self.divideProcessorsMetricProportional(clusters, trajToMetricProportional, spawningParams)

        self.degeneracyTotal = np.array(self.degeneracyInverselyProportional) + np.array(self.degeneracyMetricProportional)
        return self.degeneracyTotal.tolist()

    def divideProcessorsMetricProportional(self, clusters, trajToDistribute, spawningParams):
        """
            Distribute the trajectories among the clusters according to their
            metric.

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param weights: Weight of each cluster
            :type weights: np.Array
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int

            :returns: list -- List with the number of processors allocated to
                each cluster
        """

        metrics = self.getMetrics(clusters)

        if isinstance(metrics, list):
            metrics = np.array(metrics)

        # Shift so that differences become larger.
        # Also, we can now merge positive & negative values
        # Alternatives: Boltzmann weights
        if spawningParams.condition == blockNames.SpawningParams.minValue:
            shiftValue = np.max(metrics)
        else:
            shiftValue = np.min(metrics)
        shiftedMetrics = np.subtract(metrics, shiftValue)
        bestClusters = shiftedMetrics.argsort()

        if spawningParams.condition == blockNames.SpawningParams.minValue:
            shiftedMetrics[bestClusters[spawningParams.nclusters:]] = 0  # only consider best ones
        else:
            shiftedMetrics[bestClusters[:-spawningParams.nclusters]] = 0  # only consider best ones

        metricWeights = spawningParams.metricWeights
        if metricWeights == blockNames.SpawningParams.linear:

            # all shiftedMetrics <= 0, sum(shiftedMetrics) < 0 => weights >= 0
            if abs(shiftedMetrics.sum()) < 1e-8:
                weights = np.ones(len(metrics))/len(metrics)
            else:
                weights = (1.*shiftedMetrics)/sum(shiftedMetrics)

        elif metricWeights == blockNames.SpawningParams.boltzmann:
            T = spawningParams.temperature
            kbT = 0.001987*T
            if abs(shiftedMetrics.sum()) < 1e-8:
                weights = np.ones(len(metrics))/len(metrics)
            else:
                weights = np.exp(-shiftedMetrics/kbT)
                weights /= sum(weights)
        else:
            raise ValueError("No appropiate value for the metricWeights "
                             "was found, please specify a correct value. The "
                             "default value of the metrics weighting is linear")

        return self.divideTrajAccordingToWeights(weights, trajToDistribute)


class VariableEpsilonDegeneracyCalculator(DensitySpawningCalculator):

    def __init__(self, densityCalculator=densitycalculator.NullDensityCalculator()):
        DensitySpawningCalculator.__init__(self, densityCalculator)
        self.epsilonDegeneracyCalculator = EpsilonDegeneracyCalculator(densityCalculator)
        self.type = spawningTypes.SPAWNING_TYPES.variableEpsilon
        self.degeneracyInverselyProportional = None
        self.degeneracyMetricProportional = None
        self.degeneracyTotal = None
        self.maxContacts = None
        # print variable epsilon information
        epsilon_file = open("epsilon_values.txt", "w")
        epsilon_file.write("Iteration\tEpsilon\n")
        epsilon_file.close()

    def linearVariation(self, spawningParams, currentEpoch):
        """
            Calculate linear variation of epsilon with respect to the iteraetion

            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int
        """
        if currentEpoch == 0:
            spawningParams.epsilon = spawningParams.minEpsilon
            return

        middleWindow = int(spawningParams.period/2)
        leftWindow = int(spawningParams.maxEpsilonWindow/2)
        rightWindow = leftWindow+middleWindow
        if currentEpoch == spawningParams.period-1:
            # Avoid negative epsilon for minEpsilon = 0
            return
        rateEpsilonVariation = [(spawningParams.maxEpsilon-spawningParams.minEpsilon)/(middleWindow-leftWindow), (spawningParams.maxEpsilon-spawningParams.minEpsilon)/(spawningParams.period-rightWindow-1)]
        spawningParams.epsilon += return_sign(currentEpoch, leftWindow,
                                              middleWindow, rightWindow) * rateEpsilonVariation[currentEpoch > middleWindow]

    def contactsVariation(self, clusters, spawningParams):
        """
            Calculate the variation of epsilon according to the contacts ratio

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
        """
        if self.maxContacts is None:
            self.maxContacts = reduce(max, [cluster.contacts for cluster in clusters])
        maxContacts = reduce(max, [cluster.contacts for cluster in clusters])
        if spawningParams.epsilon < spawningParams.maxEpsilon:
            spawningParams.epsilon += calculateContactsVar(maxContacts-self.maxContacts, spawningParams.maxEpsilon)
        self.maxContacts = maxContacts

    def calculateEpsilonValue(self, spawningParams, currentEpoch, clusters):
        """
            Calculate variation of epsilon according to the selected parameters

            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int
            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
        """
        if spawningParams.varEpsilonType == blockNames.VariableEpsilonTypes.linearVariation:
            if currentEpoch is None or spawningParams.variationWindow < currentEpoch:
                spawningParams.epsilon = spawningParams.minEpsilon
                return
            self.linearVariation(spawningParams,
                                 (currentEpoch % spawningParams.period))
        elif spawningParams.varEpsilonType == blockNames.VariableEpsilonTypes.contactsVariation:
            self.contactsVariation(clusters, spawningParams)
        else:
            sys.exit("Unknown epsilon variation type! Choices are: " +
                     str(spawningTypes.EPSILON_VARIATION_TYPE_TO_STRING_DICTIONARY.values()))

    def logVariableEpsilon(self, epsilon, epoch):
        """
            Log spawning information
        """
        with open("epsilon_values.txt", "a") as epsilon_file:
            epsilon_file.write("%d\t%f\n" % (epoch, epsilon))

    def calculate(self, clusters, trajToDistribute, spawningParams, currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        self.calculateEpsilonValue(spawningParams, currentEpoch, clusters)
        self.logVariableEpsilon(spawningParams.epsilon, currentEpoch)
        return self.epsilonDegeneracyCalculator.calculate(clusters, trajToDistribute, spawningParams, currentEpoch)


class SimulatedAnnealingCalculator(SpawningCalculator):

    def __init__(self):
        SpawningCalculator.__init__(self)
        self.type = spawningTypes.SPAWNING_TYPES.simulatedAnnealing

    def log(self):
        """
            Log spawning information
        """
        pass

    def computeTemperature(self, params, epoch):
        T = params.temperature - params.decrement*epoch
        if T < 300:
            return 300
        else:
            return T

    def calculate(self, clusters, trajToDistribute, spawningParams,
                  currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        metrics = self.getMetrics(clusters)

        minimumValue = np.min(metrics)
        shiftedMetrics = np.subtract(metrics, minimumValue)

        T = self.computeTemperature(spawningParams, currentEpoch)
        kbT = 0.001987*T
        weights = np.exp(-shiftedMetrics/kbT)
        weights /= sum(weights)

        return self.divideProportionalToArray(weights, trajToDistribute)


class FASTDegeneracyCalculator(DensitySpawningCalculator):

    def __init__(self, densityCalculator=densitycalculator.NullDensityCalculator()):
        DensitySpawningCalculator.__init__(self, densityCalculator)
        self.type = spawningTypes.SPAWNING_TYPES.FAST
        self.densityCalculator = densityCalculator

    def normaliseArray(self, array):
        maxValue = float(np.max(array))
        minValue = float(np.min(array))

        if maxValue-minValue > 1e-5:
            normalisedArray = np.subtract(maxValue, array)/(maxValue - minValue)
        else:
            # all elements are equal
            n = len(array)
            normalisedArray = np.ones(n)/float(n)
        return normalisedArray

    def calculateNormalisedSizes(self, clusters):
        sizes = getSizes(clusters)

        densities = self.calculateDensities(clusters)
        weightedSizes = sizes/densities

        return self.normaliseArray(weightedSizes)

    def calculateNormalisedMetrics(self, clusters):
        metrics = self.getMetrics(clusters)
        return self.normaliseArray(metrics)

    def calculate(self, clusters, trajToDivide, spawningParams, currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        normalisedSizes = self.calculateNormalisedSizes(clusters)
        normalisedMetrics = self.calculateNormalisedMetrics(clusters)

        weight = normalisedSizes + 1.*normalisedMetrics

        return self.divideProportionalToArray(weight, trajToDivide)

    def log(self):
        """
            Log spawning information
        """
        pass


class UCBCalculator(DensitySpawningCalculator):

    def __init__(self, densityCalculator=densitycalculator.NullDensityCalculator()):
        DensitySpawningCalculator.__init__(self, densityCalculator)
        self.type = spawningTypes.SPAWNING_TYPES.UCB
        self.prevMetrics = np.array([0.0])
        self.averages = []
        self.beta = 1.0
        self.averageMetric = 0
        self.epoch = np.array([0.0])

    def log(self):
        """
            Log spawning information
        """
        pass

    def calculate(self, clusters, trajToDistribute, spawningParams, currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        self.epoch += 1
        sizes = np.array(getSizes(clusters))
        densities = self.calculateDensities(clusters)
        metrics = np.array(self.getMetrics(clusters))
        # self.averageMetric = np.mean(metrics)
        maximumValue = np.max(metrics)
        shiftedMetrics = np.subtract(metrics, maximumValue)
        if abs(shiftedMetrics.sum()) < 1e-8:
            weights = np.ones(len(metrics))/len(metrics)
        else:
            weights = (1.*shiftedMetrics)/min(shiftedMetrics)
        # for i, metric in enumerate(metrics):
        #     if i < len(self.prevMetrics):
        #         self.prevMetrics[i] = abs(metric-self.prevMetrics[i])/abs(metric)
        #         self.averages[i] += (self.prevMetrics[i]-self.averages[i])/sizes[i]
        #     else:
        #         self.prevMetrics.append(-(metric-self.averageMetric)/abs(metric))
        #         self.averages.append(-(metric-self.averageMetric)/abs(metric))
        l = self.prevMetrics.size
        n = weights.size
        try:
            self.prevMetrics = np.pad(self.prevMetrics, (0, n-l), str('constant'), constant_values=(0.0))
            self.epoch = np.pad(self.epoch, (0, n-l), str('constant'), constant_values=(1.0))
        except AttributeError:
            # Numpy version in life is too old to use pad function
            prevMetrics = np.zeros_like(weights)
            epochs = np.ones_like(weights)
            prevMetrics[:l] = self.prevMetrics
            self.prevMetrics = prevMetrics
            epochs[:l] = self.epoch
            self.epoch = epochs
        # avg[:l] = self.prevMetrics[:l]
        self.prevMetrics += (weights-self.prevMetrics)/self.epoch
        argweights = self.prevMetrics.argsort()
        weights_trimmed = np.zeros(len(sizes))
        weights_trimmed[argweights[-trajToDistribute:]] = self.prevMetrics[argweights[-trajToDistribute:]]
        # values = weights_trimmed+spawningParams.alpha*np.sqrt((1/sizes))
        # values = self.beta*weights_trimmed**2+spawningParams.alpha*(1/sizes**2)
        values = self.beta*weights_trimmed**2+spawningParams.alpha*(1/sizes)
        # values = self.beta*weights_trimmed**2+(spawningParams.alpha/((np.log2(currentEpoch+2))**(1/4.0)))*(1/sizes)
        # minVal = np.min(values)
        # if minVal < 0:
        #     # if there is a negative value shift all the values so that the min
        #     # value is zero
        #    values += abs(minVal)
        if densities.any():
            weights = values*densities
        else:
            weights = values
        argweights = weights.argsort()
        weights_trimmed = np.zeros(len(sizes))
        weights_trimmed[argweights[-trajToDistribute:]] = weights[argweights[-trajToDistribute:]]
        return self.divideProportionalToArray(weights_trimmed, trajToDistribute)


class REAPCalculator(DensitySpawningCalculator):
    def __init__(self, densityCalculator=densitycalculator.NullDensityCalculator()):
        """
            Spawning following the Reinforcement learning based Adaptive
            samPling (REAP) (Shamsi et al., arXiv, Oct 2017), where the reward
            given by the exploration on several reaction coordinates
            is maximized
        """
        DensitySpawningCalculator.__init__(self, densityCalculator)
        self.type = spawningTypes.SPAWNING_TYPES.REAP
        self.weights = None
        self.metricInd = None
        self.rewards = None
        self.degeneracy = None
        # constraints so the weights have values between 0 and 1
        self.cons = ({'type': 'eq', 'fun': lambda x: np.array(x.sum()-1)})
        self.bounds = None

    def calculate(self, clusters, trajToDivide, spawningParams, currentEpoch=None):
        """
            Calculate the degeneracy of the clusters

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: list -- List containing the degeneracy of the clusters
        """
        population = []
        metrics = []
        if self.metricInd is None:
            if spawningParams.metricInd == -1:
                self.metricInd = list(range(3, clusters[0].metrics.size))
            else:
                self.metricInd = spawningParams.metricInd
            self.bounds = [(0, 1)]*len(self.metricInd)

        # Gather population and metrics data for all clusters
        for cluster in clusters:
            population.append(float(cluster.elements))
            metrics.append([cluster.metrics[i] for i in self.metricInd])
        self.degeneracy = np.zeros_like(population)
        metrics = np.array(metrics).T
        meanRew = np.mean(metrics, axis=1)
        stdRew = np.std(metrics, axis=1)
        # Filter top least populated clusters
        population = np.array(population)
        densities = self.calculateDensities(clusters)
        if densities.any():
            population /= densities
        argweights = np.argsort(population)
        metrics = metrics[:, argweights[:trajToDivide]]
        # Shift and scale all metrics to have mean 0 and std 1, so that the
        # weight of each metric is not affected by its magnitude (i.e. binding
        # energy ~ 10**2 while SASA <= 1)
        rewProv = np.abs(metrics-meanRew[:, np.newaxis])/stdRew[:, np.newaxis]

        if self.weights is None:
            self.weights = np.ones(len(self.metricInd))/len(self.metricInd)
        else:
            optimResult = optim.minimize(reward, self.weights, args=(rewProv,),
                                         method="SLSQP", constraints=self.cons,
                                         bounds=self.bounds)
            self.weights = optimResult.x
        self.rewards = (self.weights[:, np.newaxis]*rewProv).sum(axis=0)
        self.degeneracy[argweights[:trajToDivide]] = self.divideProportionalToArray(self.rewards, trajToDivide)
        return self.degeneracy.tolist()

    def log(self):
        """
            Log spawning information
        """
        if self.degeneracy is not None:
            print("[SpawningLog] Total: %s" % str(self.degeneracy))
        print("Metric indices")
        print(self.metricInd)
        print("Spawning weights")
        print(self.weights)


class NullSpawningCalculator(SpawningCalculator):
    def __init__(self):
        SpawningCalculator.__init__(self)
        self.type = spawningTypes.SPAWNING_TYPES.null

    def calculate(self, clusters, trajToDivide, spawningParams, currentEpoch=None):
        """
            Calculate the degeneracy of the clusters. In this particular class
            no spawning is performed, so this function just returns None

            :param clusters: Existing clusters
            :type clusters: :py:class:`.Clusters`
            :param trajToDistribute: Number of processors to distribute
            :type trajToDistribute: int
            :param spawningParams: Object containing the parameters of the spawning
            :type spawningParams: :py:class:`.SpawningParams`
            :param currentEpoch: Current iteration number
            :type currentEpoch: int

            :returns: None
        """
        return None



from __future__ import absolute_import, division, print_function, unicode_literals
from io import open
from AdaptivePELE.atomset import RMSDCalculator
from AdaptivePELE.atomset import SymmetryContactMapEvaluator as sym
from AdaptivePELE.clustering import clustering
import AdaptivePELE.atomset.atomset as atomset
import unittest
import numpy as np


class atomsetTest(unittest.TestCase):
    """ For the moment the tests include loading from file and string, resname
    and atomname selection. It uses a toy pdb of only 5 lines located in
    tests/data
   TODO: heavyAtoms and type selection
    """
    def testPDB_from_file(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test.pdb")

        # assertion
        pdbContent = "ATOM      1  N   ASN A   1       7.920  22.268   9.257  1.00 15.18           N1+\n\
ATOM      2  CA  CYS A   1       8.394  20.916   9.575  1.00 16.24           C  \n\
ATOM      3  CA  ASN A   1       7.870  19.937   8.524  1.00 16.63           C  \n\
ATOM      4  O   ASN A   1       7.030  20.308   7.700  1.00 16.13           O  \n\
ATOM      5  CB  CYS A   1       8.108  20.445  11.030  1.00 16.53           C  \n"
        atom1 = atomset.Atom("ATOM      1  N   ASN A   1       7.920  22.268   9.257 1.00 15.18           N1+")
        atom2 = atomset.Atom("ATOM      2  CA  CYS A   1       8.394  20.916   9.575 1.00 16.24           C  ")
        atom3 = atomset.Atom("ATOM      3  CA  ASN A   1       7.870  19.937   8.524 1.00 16.63           C  ")
        atom4 = atomset.Atom("ATOM      4  O   ASN A   1       7.030  20.308   7.700 1.00 16.13           O  ")
        atom5 = atomset.Atom("ATOM      5  CB  CYS A   1       8.108  20.445  11.030 1.00 16.53           C  ")
        goldenAtomsDict = {atom1.id: atom1, atom2.id: atom2, atom3.id: atom3, atom4.id: atom4, atom5.id: atom5}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_from_str(self):
        # preparation
        with open("tests/data/pdb_test.pdb", "rt") as pdbfile:
            pdbstring = pdbfile.read()
        pdb = atomset.PDB()

        # function to test
        pdb.initialise(pdbstring)

        # assertion
        pdbContent = "ATOM      1  N   ASN A   1       7.920  22.268   9.257  1.00 15.18           N1+\n\
ATOM      2  CA  CYS A   1       8.394  20.916   9.575  1.00 16.24           C  \n\
ATOM      3  CA  ASN A   1       7.870  19.937   8.524  1.00 16.63           C  \n\
ATOM      4  O   ASN A   1       7.030  20.308   7.700  1.00 16.13           O  \n\
ATOM      5  CB  CYS A   1       8.108  20.445  11.030  1.00 16.53           C  \n"
        atom1 = atomset.Atom("ATOM      1  N   ASN A   1       7.920  22.268   9.257 1.00 15.18           N1+")
        atom2 = atomset.Atom("ATOM      2  CA  CYS A   1       8.394  20.916   9.575 1.00 16.24           C  ")
        atom3 = atomset.Atom("ATOM      3  CA  ASN A   1       7.870  19.937   8.524 1.00 16.63           C  ")
        atom4 = atomset.Atom("ATOM      4  O   ASN A   1       7.030  20.308   7.700 1.00 16.13           O  ")
        atom5 = atomset.Atom("ATOM      5  CB  CYS A   1       8.108  20.445  11.030 1.00 16.53           C  ")
        goldenAtomsDict = {atom1.id: atom1, atom2.id: atom2, atom3.id: atom3, atom4.id: atom4, atom5.id: atom5}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_sel_resname(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test.pdb", resname="CYS")

        # assertion
        pdbContent = "ATOM      1  N   ASN A   1       7.920  22.268   9.257  1.00 15.18           N1+\n\
ATOM      2  CA  CYS A   1       8.394  20.916   9.575  1.00 16.24           C  \n\
ATOM      3  CA  ASN A   1       7.870  19.937   8.524  1.00 16.63           C  \n\
ATOM      4  O   ASN A   1       7.030  20.308   7.700  1.00 16.13           O  \n\
ATOM      5  CB  CYS A   1       8.108  20.445  11.030  1.00 16.53           C  \n"
        atom2 = atomset.Atom("ATOM      2  CA  CYS A   1       8.394  20.916   9.575 1.00 16.24           C  ")
        atom5 = atomset.Atom("ATOM      5  CB  CYS A   1       8.108  20.445  11.030 1.00 16.53           C  ")
        goldenAtomsDict = {atom2.id: atom2, atom5.id: atom5}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_sel_atomname(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test.pdb", atomname="CA")

        # assertion
        pdbContent = "ATOM      1  N   ASN A   1       7.920  22.268   9.257  1.00 15.18           N1+\n\
ATOM      2  CA  CYS A   1       8.394  20.916   9.575  1.00 16.24           C  \n\
ATOM      3  CA  ASN A   1       7.870  19.937   8.524  1.00 16.63           C  \n\
ATOM      4  O   ASN A   1       7.030  20.308   7.700  1.00 16.13           O  \n\
ATOM      5  CB  CYS A   1       8.108  20.445  11.030  1.00 16.53           C  \n"
        atom2 = atomset.Atom("ATOM      2  CA  CYS A   1       8.394  20.916   9.575 1.00 16.24           C  ")
        atom3 = atomset.Atom("ATOM      3  CA  ASN A   1       7.870  19.937   8.524 1.00 16.63           C  ")
        goldenAtomsDict = {atom2.id: atom2, atom3.id: atom3}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_sel_type_protein(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test_ligand.pdb",
                       type="PROTEIN")

        # assertion
        pdbContent = "MODEL        1\n\
ATOM   1717  H   ASN A 119      25.915   9.925  -7.236  1.00 31.61           H  \n\
ATOM   1718  CA  ASN A 119      27.159  10.509  -6.736  1.00 33.83           C  \n\
TER\n\
HETATM 1733  O1  AIN L   1      13.907  16.130   0.624  0.50 28.52           O  \n\
TER\n\
HETATM 1753 CA    CA B   1      16.636  15.477   0.293  1.00  3.39          Ca2+\n\
TER\n\
ENDMDL\n\
END   \n"

        atom1 = atomset.Atom("ATOM   1718  CA  ASN A 119      27.159  10.509  -6.736  1.00 33.83           C  ")
        goldenAtomsDict = {atom1.id: atom1}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_sel_type_hetero(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test_ligand.pdb", type="HETERO")

        # assertion
        pdbContent = "MODEL        1\n\
ATOM   1717  H   ASN A 119      25.915   9.925  -7.236  1.00 31.61           H  \n\
ATOM   1718  CA  ASN A 119      27.159  10.509  -6.736  1.00 33.83           C  \n\
TER\n\
HETATM 1733  O1  AIN L   1      13.907  16.130   0.624  0.50 28.52           O  \n\
TER\n\
HETATM 1753 CA    CA B   1      16.636  15.477   0.293  1.00  3.39          Ca2+\n\
TER\n\
ENDMDL\n\
END   \n"

        atom1 = atomset.Atom("HETATM 1733  O1  AIN L   1      13.907  16.130   0.624  0.50 28.52           O  ")
        atom2 = atomset.Atom("HETATM 1753 CA    CA B   1      16.636  15.477   0.293  1.00  3.39          Ca2+")
        goldenAtomsDict = {atom1.id: atom1, atom2.id: atom2}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_sel_type_heavyAtoms(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test_ligand.pdb", heavyAtoms=False)

        # assertion
        pdbContent = "MODEL        1\n\
ATOM   1717  H   ASN A 119      25.915   9.925  -7.236  1.00 31.61           H  \n\
ATOM   1718  CA  ASN A 119      27.159  10.509  -6.736  1.00 33.83           C  \n\
TER\n\
HETATM 1733  O1  AIN L   1      13.907  16.130   0.624  0.50 28.52           O  \n\
TER\n\
HETATM 1753 CA    CA B   1      16.636  15.477   0.293  1.00  3.39          Ca2+\n\
TER\n\
ENDMDL\n\
END   \n"

        atom1 = atomset.Atom("ATOM   1717  H   ASN A 119      25.915   9.925  -7.236  1.00 31.61           H  ")
        atom2 = atomset.Atom("ATOM   1718  CA  ASN A 119      27.159  10.509  -6.736  1.00 33.83           C  ")
        atom3 = atomset.Atom("HETATM 1733  O1  AIN L   1      13.907  16.130   0.624  0.50 28.52           O  ")
        atom4 = atomset.Atom("HETATM 1753 CA    CA B   1      16.636  15.477   0.293  1.00  3.39          Ca2+")
        goldenAtomsDict = {atom1.id: atom1, atom2.id: atom2, atom3.id: atom3, atom4.id: atom4}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_sel_resnum(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test2.pdb", resnum=2)

        # assertion
        pdbContent = "ATOM      1  N   ASN A   1       7.920  22.268   9.257  1.00 15.18           N1+\n\
ATOM      2  CA  CYS B   1       8.394  20.916   9.575  1.00 16.24           C  \n\
ATOM      3  CA  ASN B   1       7.870  19.937   8.524  1.00 16.63           C  \n\
ATOM      4  O   ASN A   2       7.030  20.308   7.700  1.00 16.13           O  \n\
ATOM      5  CB  CYS A   2       8.108  20.445  11.030  1.00 16.53           C  \n"
        atom2 = atomset.Atom("ATOM      4  O   ASN A   2       7.030  20.308   7.700  1.00 16.13           O  ")
        atom3 = atomset.Atom("ATOM      5  CB  CYS A   2       8.108  20.445  11.030  1.00 16.53           C  ")
        goldenAtomsDict = {atom2.id: atom2, atom3.id: atom3}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_sel_chain(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test2.pdb", chain="B")

        # assertion
        pdbContent = "ATOM      1  N   ASN A   1       7.920  22.268   9.257  1.00 15.18           N1+\n\
ATOM      2  CA  CYS B   1       8.394  20.916   9.575  1.00 16.24           C  \n\
ATOM      3  CA  ASN B   1       7.870  19.937   8.524  1.00 16.63           C  \n\
ATOM      4  O   ASN A   2       7.030  20.308   7.700  1.00 16.13           O  \n\
ATOM      5  CB  CYS A   2       8.108  20.445  11.030  1.00 16.53           C  \n"

        atom2 = atomset.Atom("ATOM      2  CA  CYS B   1       8.394  20.916   9.575 1.00 16.24           C  ")
        atom3 = atomset.Atom("ATOM      3  CA  ASN B   1       7.870  19.937   8.524 1.00 16.63           C  ")
        goldenAtomsDict = {atom2.id: atom2, atom3.id: atom3}

        self.assertEqual(pdb.pdb, pdbContent)
        self.assertEqual(pdb.atoms, goldenAtomsDict)

    def testPDB_COM(self):
        # preparation
        pdb = atomset.PDB()

        # function to test
        pdb.initialise("tests/data/pdb_test.pdb")

        # assertion
        total_mass = 66.0382
        COM_array = np.array([516.1336264, 1373.048894, 602.7150822])
        pdb.extractCOM()
        self.assertNotAlmostEqual(pdb.totalMass, 0)
        COM = COM_array/pdb.totalMass

        np.testing.assert_array_almost_equal(COM, pdb.getCOM())
        self.assertAlmostEqual(total_mass, pdb.totalMass)

    def testPDB_write(self):
        # preparation
        pdb = atomset.PDB()
        pdb.initialise("tests/data/pdb_test.pdb")

        # function to test
        pdb.writePDB("tests/data/pdb_test_write.pdb")

        # assertion
        pdbtestfile = open("tests/data/pdb_test_write.pdb", "r")
        pdbtestsstr = pdbtestfile.read()
        pdbtestfile.close()
        self.assertEqual(pdb.pdb, pdbtestsstr)

    def testPDB_RMSD(self):
        # preparation
        pdb_native = atomset.PDB()
        pdb_native.initialise("tests/data/ain_native_fixed.pdb", resname='AIN')
        pdb_traj = atomset.PDB()
        pdb_traj.initialise("tests/data/ain_trajectory.pdb", resname='AIN')
        RMSDCalc = RMSDCalculator.RMSDCalculator()

        # function to test
        RMSD = RMSDCalc.computeRMSD(pdb_native, pdb_traj)
        golden_RMSD = 3.928617
        self.assertAlmostEqual(RMSD, golden_RMSD, 5)

    def testPDB_RMSD_symmetries(self):
        # preparation
        pdb_native = atomset.PDB()
        pdb_native.initialise("tests/data/ain_native_fixed.pdb", resname='AIN')
        pdb_traj = atomset.PDB()
        pdb_traj.initialise("tests/data/ain_trajectory.pdb", resname='AIN')
        symDict = [{"1733:O1:AIN": "1735:O2:AIN"}]
        RMSDCalc = RMSDCalculator.RMSDCalculator(symDict)
        # function to test
        RMSD = RMSDCalc.computeRMSD(pdb_native, pdb_traj)
        reverseRMSD = RMSDCalc.computeRMSD(pdb_traj, pdb_native)
        golden_RMSD = 3.860743
        self.assertAlmostEqual(RMSD, reverseRMSD, 5)
        self.assertAlmostEqual(RMSD, golden_RMSD, 5)

    def test_combination_symmetries(self):
        # preparation
        pdb_0 = atomset.PDB()
        pdb_0.initialise("tests/data/symmetries/cluster_0.pdb", resname='AEN')
        pdb_1 = atomset.PDB()
        pdb_1.initialise("tests/data/symmetries/cluster_1.pdb", resname='AEN')
        pdb_2 = atomset.PDB()
        pdb_2.initialise("tests/data/symmetries/cluster_2.pdb", resname='AEN')
        symmetries3PTB = [{"3225:C3:AEN": "3227:C5:AEN", "3224:C2:AEN": "3228:C6:AEN"},
                          {"3230:N1:AEN": "3231:N2:AEN"}]
        RMSDCalc = RMSDCalculator.RMSDCalculator(symmetries3PTB)
        # funtion to test
        RMSD02 = RMSDCalc.computeRMSD(pdb_0, pdb_2)
        RMSD20 = RMSDCalc.computeRMSD(pdb_2, pdb_0)
        RMSD01 = RMSDCalc.computeRMSD(pdb_0, pdb_1)
        RMSD10 = RMSDCalc.computeRMSD(pdb_1, pdb_0)
        RMSD21 = RMSDCalc.computeRMSD(pdb_2, pdb_1)
        RMSD12 = RMSDCalc.computeRMSD(pdb_1, pdb_2)
        self.assertEqual(RMSD01, RMSD10)
        self.assertEqual(RMSD02, RMSD20)
        self.assertEqual(RMSD21, RMSD12)

    def testPDB_contacts(self):
        # preparation
        pdb_native = atomset.PDB()
        pdb_native.initialise("tests/data/native_ain.pdb")

        # function to test
        contacts = pdb_native.countContacts("AIN", 8)
        golden_contacts = 19
        self.assertEqual(contacts, golden_contacts)

    def testPDB_contactmap(self):
        # preparation
        pdb_native = atomset.PDB()
        pdb_native.initialise("tests/data/pdb_test_contact.pdb")
        symmetryEvaluator = sym.SymmetryContactMapEvaluator([])

        # function to test
        contact_map, contacts = symmetryEvaluator.createContactMap(pdb_native,
                                                                   "AIN", 8)
        golden_contact_map = np.array([[1, 0, 0, 0], [0, 1, 1, 1]])
        golden_contacts = pdb_native.countContacts("AIN", 8)
        np.testing.assert_array_equal(contact_map, golden_contact_map)
        self.assertEqual(golden_contacts, contacts)

    def test_contactMapContacts(self):
        # preparation
        pdb_1 = atomset.PDB()
        pdb_1.initialise("tests/data/pdb_test_contact.pdb", resname='AIN')
        symmetryEvaluator = sym.SymmetryContactMapEvaluator([])

        # function to test
        _, contacts = symmetryEvaluator.createContactMap(pdb_1, "AIN", 16)
        golden_contacts = pdb_1.countContacts("AIN", 8)
        self.assertEqual(golden_contacts, contacts)

    def test_symmetryContactMapJaccard(self):
        pdb_1 = atomset.PDB()
        pdb_1.initialise("tests/data/symmetries/cluster_1.pdb", resname='AEN')
        pdb_1_sym = atomset.PDB()
        pdb_1_sym.initialise("tests/data/symmetries/cluster_1_sym.pdb",
                             resname='AEN')
        symmetries3PTB = [{"3230:N1:AEN": "3231:N2:AEN"}]
        symmetryEvaluator = sym.SymmetryContactMapEvaluator(symmetries3PTB)
        symmetryEvaluatorEmpty = sym.SymmetryContactMapEvaluator()

        contactMap1, contacts1 = symmetryEvaluator.buildContactMap(pdb_1, 'AEN', 16)
        cluster = clustering.Cluster(pdb_1, contactMap=contactMap1)
        contactMap1Sym, contactsSym = symmetryEvaluator.createContactMap(pdb_1_sym, 'AEN', 16)
        contactMapNoSym, _ = symmetryEvaluator.createContactMap(pdb_1_sym, 'AEN', 16)

        goldenJaccard = 0.0
        Jaccard = symmetryEvaluator.evaluateJaccard(contactMap1Sym, cluster.contactMap)
        JaccardNosym = symmetryEvaluatorEmpty.evaluateJaccard(contactMapNoSym, cluster.contactMap)

        self.assertEqual(contacts1, contactsSym)
        self.assertAlmostEqual(goldenJaccard, Jaccard)
        self.assertNotAlmostEqual(Jaccard, JaccardNosym)

    def test_symmetryContactMapCorrelation(self):
        pdb_1 = atomset.PDB()
        pdb_1.initialise("tests/data/symmetries/cluster_1.pdb", resname='AEN')
        pdb_1_sym = atomset.PDB()
        pdb_1_sym.initialise("tests/data/symmetries/cluster_1_sym.pdb",
                             resname='AEN')
        symmetries3PTB = [{"3230:N1:AEN": "3231:N2:AEN"}]
        symmetryEvaluator = sym.SymmetryContactMapEvaluator(symmetries3PTB)
        symmetryEvaluatorEmpty = sym.SymmetryContactMapEvaluator()

        contactMap1, contacts1 = symmetryEvaluator.buildContactMap(pdb_1, 'AEN', 16)
        cluster = clustering.Cluster(pdb_1, contactMap=contactMap1)
        contactMap1Sym, contactsSym = symmetryEvaluator.createContactMap(pdb_1_sym, 'AEN', 16)
        contactMapNoSym, _ = symmetryEvaluator.createContactMap(pdb_1_sym, 'AEN', 16)

        goldenCorrelation = 0.0
        correlationSym = symmetryEvaluator.evaluateCorrelation(contactMap1Sym, cluster.contactMap)
        correlationNosym = symmetryEvaluatorEmpty.evaluateCorrelation(contactMapNoSym, cluster.contactMap)

        self.assertEqual(contacts1, contactsSym)
        self.assertAlmostEqual(goldenCorrelation, correlationSym)
        self.assertNotAlmostEqual(correlationSym, correlationNosym)

    def test_symmetryContactMapDifference(self):
        pdb_1 = atomset.PDB()
        pdb_1.initialise("tests/data/symmetries/cluster_1.pdb", resname='AEN')
        pdb_1_sym = atomset.PDB()
        pdb_1_sym.initialise("tests/data/symmetries/cluster_1_sym.pdb",
                             resname='AEN')
        symmetries3PTB = [{"3230:N1:AEN": "3231:N2:AEN"}]
        symmetryEvaluator = sym.SymmetryContactMapEvaluator(symmetries3PTB)
        symmetryEvaluatorEmpty = sym.SymmetryContactMapEvaluator()

        contactMap1, contacts1 = symmetryEvaluator.buildContactMap(pdb_1, 'AEN', 16)
        cluster = clustering.Cluster(pdb_1, contactMap=contactMap1)
        contactMap1Sym, contactsSym = symmetryEvaluator.createContactMap(pdb_1_sym, 'AEN', 16)
        contactMapNoSym, _ = symmetryEvaluator.createContactMap(pdb_1_sym, 'AEN', 16)

        goldenDifference = 0.0
        DifferenceSym = symmetryEvaluator.evaluateDifferenceDistance(contactMap1Sym, cluster.contactMap)
        DifferenceNosym = symmetryEvaluatorEmpty.evaluateDifferenceDistance(contactMapNoSym, cluster.contactMap)

        self.assertEqual(contacts1, contactsSym)
        self.assertAlmostEqual(goldenDifference, DifferenceSym)
        self.assertNotAlmostEqual(DifferenceSym, DifferenceNosym)

    def test_PDB_interface(self):
        pdb = atomset.PDB()
        pdb.initialise("tests/data/symmetries/cluster_1.pdb", resname='AEN')
        self.assertEqual(len(pdb), 9)
        atomList = [atom for atom in pdb]
        atoms = [pdb.getAtom(a) for a in pdb.atomList]
        self.assertEqual(atomList, atoms)
        atomId = pdb.atomList[0]
        atom = pdb[atomId]
        self.assertEqual(atom, pdb.getAtom(atomId))
        pdb[atomId] = None
        self.assertEqual(None, pdb.getAtom(atomId))



import sys
import logging
import glob
import prody
import frag_pele.Growing.add_fragment_from_pdbs as addfr


# Getting the name of the module for the log system
logger = logging.getLogger(__name__)


def check_duplicated_pdbatomnames(pdb_content):
    """
    It checks if the content of a PDB file contains repeated PDB atom names.
    :param pdb_content: string with the content of a PDB file.
    :return: if repeated atom names: exit and complain.
    """
    pdb_atom_names_list = []
    for line in pdb_content:
        if line.startswith("HETATM") and line[17:30] != "HOH" and line[21:22] == "L":
            pdb_atom_name = line[12:16]
            pdb_atom_names_list.append(pdb_atom_name)
    set_to_check = set(pdb_atom_names_list)
    list_to_check = sorted(list(set_to_check))
    sorted_list_names = sorted(pdb_atom_names_list)
    if list_to_check != sorted_list_names:
        print(list_to_check)
        print(sorted_list_names)
        sys.exit("SOME REPEATED PDB ATOM NAMES of the ligand IN PDB FILES!!")


def check_and_fix_pdbatomnames(pdb_file):
    """
    It checks if atoms of the ligand of a PDB file contains the character 'G' (usually added by FrAG to identify atoms
    that have been grown) and modify the name of these atoms adding it element symbol to the PDB atom name.
    :param pdb_file: PDB file. str
    :return: it rewrites the PDB file applying the modifications.
    """
    with open(pdb_file) as pdb:
        content = pdb.readlines()
        check_duplicated_pdbatomnames(content)
        for i, line in enumerate(content):
            if line.startswith("HETATM") and line[21:22] == "L":
                atom_name = line[12:16]
                if atom_name.strip().startswith("G"):
                    new_atom_name = line[77:78] + atom_name.strip()
                    line_to_list = list(line)
                    line_to_list[12:16] = new_atom_name + " " * (4-len(new_atom_name))
                    line_to_list = "".join(line_to_list)
                    content[i] = line_to_list
        check_duplicated_pdbatomnames(content)
        new_pdb = "".join(content)
    with open(pdb_file, "w") as writepdb:
        writepdb.write("{}".format(new_pdb))


def check_if_atom_exists_in_ligand(pdb_file, atom_name, ligand_chain="L"):
    """
    It checks if an atom is found in a certain PDB file.
    :param pdb_file: PDB file. str
    :param atom_name: PDB atom name. str(len <= 4)
    :return: if the atom is found it prints a text and if not raise an exception.
    """
    try:
        ligand = addfr.extract_heteroatoms_pdbs(pdb_file, create_file=False, ligand_chain=ligand_chain, get_ligand=True)
    except OSError:
        raise OSError("Check filepath {} exists".format(pdb_file))
    atom = ligand.select("name {}".format(atom_name))
    try:
        print("PDB ATOM NAME selected: {} in {}.".format(atom.getNames(), pdb_file))
    except Exception as e:
        logger.critical(e)
        sys.exit("Check if the selected atom '{}' exists in '{}'".format(atom_name, pdb_file))



"""
    Create two PNG images with RMSD-MC steps and BE-RMSD for a set of different
    adaptive runs
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import subprocess
import glob
import argparse
from AdaptivePELE.analysis import plotAdaptive


def parseArguments():
    """
        Parse command line arguments
    """
    desc = "It makes two PNG images with RMSD-steps and BE-RMSD for a set of different adaptive runs\n\n"\
           "Instructions:\n"\
           "-------------\n"\
           "Keys in \"folders\", \"titles\" and \"outputFilenames\" dictionaries should match (could be avoided with simple refactor...)\n"\
           "Keys are: \"#steps_#processors\", #steps is important for the plots"\
           "Change params to match simulation.\n"
    parser = argparse.ArgumentParser(description=desc, formatter_class=argparse.RawTextHelpFormatter)
    args = parser.parse_args()
    return args

if __name__ == "__main__":
    folders = {"4_512": "PRprog_4_512"}  # , "4_64":"3ptb_4_64", "4_128":"3ptb_4_128"}

    subfoldersWildcard = "inversely_*"
    subfoldersWildcard = "be_epsilon_*"

    titles = {"4_512": "n=32, 4 steps, %s"}  # , "4_64":"n=64, 4 steps, %s", "4_128":"n=128, 4 steps, %s"}

    outputFilenames = {"4_512": "512_4_%s"}  # , "4_64":"64_4_%s", "4_128":"128_4_%s"}

    parameters = {"stepsCol": 2,
                  "RMSDCol": 5,
                  "BECol": 6,
                  "reportFilename": "report_"}

    gplFolder = "/gpfs/scratch/bsc72/bsc72755/adaptiveSampling/simulation"
    tmpFolder = "/tmp"

    tmpPlotFile = os.path.join(tmpFolder, "tmp.gpl")

    gnuplot = "$SCRATCH/software/gnuplot/bin/gnuplot"

    def buildGnuplotString(title, outputFilename, params):
        gnuplotFileStringContent = """\
        set term png\n\
        set title "%(plotTitle)s"\n\
        set output "rmsd_steps_%(outputFilename)s.png"\n\
        %(rmsdStepsPringString)s\n\
        \n\
        set output "be_rmsd_%(outputFilename)s.png\n\
        %(beRmsdPrintString)s\n
        """

        stepsPerRun = params["stepsPerRun"]
        stepsCol = params["stepsCol"]
        RMSDCol = params["RMSDCol"]
        BECol = params["BECol"]
        reportFilename = params["reportFilename"]

        rmsdStepsPrintString = plotAdaptive.generatePrintString(stepsPerRun, stepsCol, RMSDCol, reportFilename, "PRINT_RMSD_STEPS")
        beRmsdPrintString = plotAdaptive.generatePrintString(stepsPerRun, RMSDCol, BECol, reportFilename, "PRINT_BE_RMSD")
        dictionary = {"plotTitle": title,
                      "outputFilename": outputFilename,
                      "rmsdStepsPringString": rmsdStepsPrintString,
                      "beRmsdPrintString": beRmsdPrintString}

        return gnuplotFileStringContent % dictionary

    parseArguments()

    for key, folder in folders.items():
        print("Folder: ", folder)
        try:
            os.chdir(folder)
        except OSError:
            continue

        parameters["stepsPerRun"] = int(key.split("_")[0])

        subfolders = glob.glob(subfoldersWildcard)
        print(subfolders)
        for subfolder in subfolders:
            os.chdir(subfolder)

            gnuplotFileContent = buildGnuplotString(titles[key] % subfolder, outputFilenames[key] % subfolder, parameters)

            with open(tmpPlotFile, "w") as f:
                f.write(gnuplotFileContent)

            proc = subprocess.Popen("%s %s" % (gnuplot, tmpPlotFile), stdout=subprocess.PIPE, shell=True)
            (out, err) = proc.communicate()
            if out:
                print(out)
            if err:
                print(err)
            os.chdir("..")

        os.chdir("..")



#!/usr/bin/python2.7

import os
import errno
import argparse
import pandas as pd
import glob
import re
import sys
import socket

"""

   Description: Parse all the reports found under 'path' and sort them all
   by the chosen criteria (Binding Energy as default) having into account the
   frequency our pele control file writes a structure through the -ofreq param
   (1 by default). To sort from higher to lower value use -f "max" otherwise
   will rank the structures from lower to higher criteria's values. The number
   of structures will be ranked is controlled by -i 'nstruct' (default 10).

   For any problem do not hesitate to contact us through the email address written below.

"""
machine = socket.getfqdn()
__author__ = "Daniel Soler Viladrich"
__email__ = "daniel.soler@nostrumbiodiscovery.com"

# DEFAULT VALUES
ORDER = "min"
CRITERIA = ["Binding", "Energy"]
OUTPUT = "Structure_{}.pdb"
N_STRUCTS = 1
FREQ = 1
REPORT = "report"
TRAJ = "trajectory"
ACCEPTED_STEPS = 'numberOfAcceptedPeleSteps'
if "bsccv" in machine:
    ACCEPTED_STEPS = 'AcceptedSteps'
DIR = os.path.abspath(os.getcwd())


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument("filename", type=str, help="Criteria we want to rank and output the strutures for. Must be a clumn of the report. i.e: Binding Energy")
    parser.add_argument("crit", type=str, nargs='+', help="Criteria we want to rank and output the strutures for. Must be a clumn of the report. i.e: Binding Energy")
    parser.add_argument("--steps", "-as", type=str, help="Name of the accepted steps column in the report files. i.e: numberOfAcceptedPeleSteps", default=ACCEPTED_STEPS)
    parser.add_argument("--path", type=str, help="Path to Pele's results root folder i.e: path=/Pele/results/", default=DIR)
    parser.add_argument("--nst", "-n", type=int, help="Number of produced structures. i.e: 20" , default=N_STRUCTS)
    parser.add_argument("--sort", "-s", type=str, help="Look for minimum or maximum value --> Options: [min/max]. i.e: max", default=ORDER)
    parser.add_argument("--ofreq", "-f", type=int, help="Every how many steps the trajectory were outputted on PELE i.e: 4", default=FREQ)
    parser.add_argument("--out", "-o", type=str, help="Output Path. i.e: BindingEnergies_apo", default="".join(CRITERIA))
    parser.add_argument("--numfolders", "-nm", action="store_true", help="Not to parse non numerical folders")
    args = parser.parse_args()

    return args.filename, os.path.abspath(args.path), " ".join(args.crit), args.nst, args.sort, args.ofreq, args.out, args.steps, args.numfolders


def main(criteria, file_name, path=DIR, n_structs=10, sort_order="min", out_freq=FREQ, output="".join(CRITERIA), steps = ACCEPTED_STEPS, numfolders=False):
    """

      Description: Rank the traj found in the report files under path
      by the chosen criteria. Finally, output the best n_structs.

      Input:

         Path: Path to look for *report* files in all its subfolders.

         Criteria: Criteria to sort the structures.
         Needs to be the name of one of the Pele's report file column.
         (Default= "Binding Energy")

         n_structs: Numbers of structures to create.

         sort_order: "min" if sorting from lower to higher "max" from high to low.

         out_freq: "Output frequency of our Pele control file"

     Output:

        f_out: Name of the n outpu
    """

    reports = glob.glob(os.path.join(path, "*report_*"))
    try:
        reports[0]
    except IndexError:
        raise IndexError("Not report file found. Check you are in adaptive's or Pele root folder")

    # Data Mining
    min_values, steps = parse_values(reports, n_structs, criteria, sort_order, steps)
    values = min_values[criteria].tolist()
    paths = min_values[DIR].tolist()
    epochs = [os.path.basename(os.path.normpath(os.path.dirname(Path))) for Path in paths]
    file_ids = min_values.report.tolist()
    step_indexes = min_values[steps].tolist()
    files_out = ["epoch{}_trajectory_{}.{}_{}{}.pdb".format(epoch, report, int(step), criteria.replace(" ",""), value) \
       for epoch, step, report, value in zip(epochs, step_indexes, file_ids, values)]
    files_out_tmp_tuple = [("epoch{}_trajectory_{}.{}_{}{}.pdb".format(epoch, report, int(step),
                                                                                criteria.replace(" ",""),
                                                                                value), value) \
       for epoch, step, report, value in zip(epochs, step_indexes, file_ids, values)]
    files_out_best = sorted(files_out_tmp_tuple, key=lambda x: x[1])[0][0]
    for f_id, f_out, step, path, n in zip(file_ids, files_out, step_indexes, paths, range(0, n_structs)):

        # Read Trajetory from PELE's output
        f_in = glob.glob(os.path.join(os.path.dirname(path), "*trajectory*_{}.pdb".format(f_id)))
        if len(f_in) == 0:
            sys.exit("Trajectory {} not found. Be aware that PELE trajectories must contain the label \'trajectory\' in their file name to be detected".format("*trajectory*_{}".format(f_id)))
        f_in = f_in[0]
        with open(f_in, 'r') as input_file:
            file_content = input_file.read()
        trajectory_selected = re.search('MODEL\s+%d(.*?)ENDMDL' %int((step)/out_freq+1), file_content,re.DOTALL)

        traj = []
        with open(os.path.join(file_name, f_out), 'w') as f:
            traj.append("MODEL     %d" %int((step)/out_freq+1))
            try:
                traj.append(trajectory_selected.group(1))
            except AttributeError:
                raise AttributeError("Model not found. Check the -f option.")
            traj.append("ENDMDL\n")
            f.write("\n".join(traj))
        print("MODEL {} has been selected".format(f_out))
    return files_out_best, files_out


def parse_values(reports, n_structs, criteria, sort_order, steps):
    """

       Description: Parse the 'reports' and create a sorted array
       of size n_structs following the criteria chosen by the user.

    """

    INITIAL_DATA = [(DIR, []),
                    (REPORT, []),
                    (steps, []),
                    (criteria, [])
                    ]
    min_values = pd.DataFrame.from_items(INITIAL_DATA)
    for f in reports:
        report_number = os.path.basename(f).split("_")[-1]
        data = pd.read_csv(f, sep='    ', engine='python')
        # Skip first line not to get initial structure
        if not steps in list(data): steps="AcceptedSteps"
        selected_data = data.loc[1:, [steps, criteria]]
        if sort_order == "min":
                report_values =  selected_data.nsmallest(n_structs, criteria)
                report_values.insert(0, DIR, [f]*report_values[criteria].size)
                report_values.insert(1, REPORT, [report_number]*report_values[criteria].size)
                mixed_values = pd.concat([min_values, report_values])
                mixed_values.index = pd.RangeIndex(len(mixed_values.index))
                min_values = mixed_values.nsmallest(n_structs, criteria)

        else:
                report_values =  selected_data.nlargest(n_structs, criteria)
                report_values.insert(0, DIR, [f]*report_values[criteria].size)
                report_values.insert(1, REPORT, [report_number]*report_values[criteria].size)
                mixed_values = pd.concat([min_values, report_values])
                min_values = mixed_values.nlargest(n_structs, criteria)
    return min_values, steps


def filter_non_numerical_folders(reports, numfolders):
    """
    Filter non numerical folders among
    the folders to parse
    """
    if(numfolders):
        new_reports = [report for report in reports if(os.path.basename(os.path.dirname(report)).isdigit())]
        return new_reports
    else:
        return reports


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


if __name__ == "__main__":
    filename, path, crit, nst, sort, ofreq, out, steps, numfolders = parse_args()
    main(crit, filename, path,  nst, sort, ofreq, out, steps, numfolders)



from AdaptivePELE.constants import blockNames

class THRESHOLD_CALCULATOR_TYPES:
    heaviside, constant = list(range(2))

THRESHOLD_CALCULATOR_TYPE_TO_STRING_DICTIONARY = {
    THRESHOLD_CALCULATOR_TYPES.heaviside : blockNames.ThresholdCalculator.heaviside,
    THRESHOLD_CALCULATOR_TYPES.constant : blockNames.ThresholdCalculator.constant
}



from __future__ import absolute_import, division, print_function, unicode_literals
import unittest
import argparse
from AdaptivePELE.tests import testSpawning as tSpawning
from AdaptivePELE.tests import testAtomset as tAtomset
from AdaptivePELE.tests import testClustering as tClustering
from AdaptivePELE.tests import testAdaptiveSampling as tAdaptive
from AdaptivePELE.tests import testThresholdcalculator as tThreshold
from AdaptivePELE.tests import testDensityCalculator as tDensity


def parse_args():
    desc = ("Run testing suite. Possible options are:\na Run all tests\n"
            "at Runs atomset tests\ns Run spawning tests\nth Run threshold "
            "calculator tests\nd Run density tests\nc Run clustering tests"
            "Ad Run adaptive integration tests")
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("--run", default=['a'], nargs="*", help="Tests to run")
    parser.add_argument("--exclude", default=[], nargs="*", help="Tests to exclude")
    args = parser.parse_args()

    return args.run, args.exclude

def main(run, exclude):
    testSuite = unittest.TestSuite()
    to_run = set(run)-set(exclude)

    if "at" in to_run or "a" in to_run:
        print("Will run atomset tests")
        testSuite.addTest(unittest.makeSuite(tAtomset.atomsetTest))
    if "s" in to_run or "a" in to_run:
        print("Will run spawning tests")
        testSuite.addTest(unittest.makeSuite(tSpawning.TestSpawningCalculator))
    if "th" in to_run or "a" in to_run:
        print("Will run threshold tests")
        testSuite.addTest(unittest.makeSuite(tThreshold.thresholdCalculatorTest))
    if "d" in to_run or "a" in to_run:
        print("Will run denstity tests")
        testSuite.addTest(unittest.makeSuite(tDensity.densityCalculatorTest))
    if "c" in to_run or "a" in to_run:
        print("Will run clustering tests")
        testSuite.addTest(unittest.makeSuite(tClustering.clusteringTest))
    if "Ad" in to_run or "a" in to_run:
        print("Will run integration tests")
        testSuite.addTest(unittest.makeSuite(tAdaptive.TestadaptiveSampling))

    runner = unittest.TextTestRunner()
    runner.run(testSuite)

if __name__ == "__main__":
    run_list, exclude_list = parse_args()
    main(run_list, exclude_list)



import numpy as np
import math
import matplotlib.pyplot as plt
from frag_pele.Growing.template_fragmenter import Atom, TemplateOPLS2005, Phi, Theta, Bond

NUM_AVOG = 6.022e23
DIEL_CNST = 1/(4*np.pi*1)
ELECTRON_CHARGE = 1.6022e-19  # C

MAGIC_NUMBER = 10e10/(4.184*1000)
Ein = 1  # ??
Esolv = 80.0  # ??
Evac = 1.257 * 10**(-6)  # H/m
I = 0.15
T = 1000  # K
R = 8.314  # J / K**-1 mol**-1


class ComputeNBONEnergies:
    def __init__(self, epsilon1=0, epsilon2=0, sigma1=0, sigma2=0, charge1=0, charge2=0, radnpSGB1=0, sgbnpType1=0,
                 radnpSGB2=0, sgbnpType2=0, radii=0):
        self.epsilon1 = epsilon1
        self.epsilon2 = epsilon2
        self.sigma1 = sigma1
        self.sigma2 = sigma2
        self.charge1 = charge1
        self.charge2 = charge2
        self.radnpSGB1 = radnpSGB1
        self.sgbnpType1 = sgbnpType1
        self.radnpSGB2 = radnpSGB2
        self.sgbnpType2 = sgbnpType2
        self.radii = radii

    def energy_vdw(self):
        energy = float(2*np.sqrt(self.epsilon1))*float(2*np.sqrt(self.epsilon2))*(((float(self.sigma1)**12) /
                 (float(self.radii)**12)) - ((float(self.sigma2)**6)/(float(self.radii)**6)))
        return energy

    def charge_function(self):
        factor = np.sqrt((DIEL_CNST)*(NUM_AVOG)*(ELECTRON_CHARGE**2)*(MAGIC_NUMBER))
        charge1 = self.charge1 * factor
        charge2 = self.charge2 * factor
        energy = (charge1*charge2)/(1*self.radii)
        return energy

    def solv_pol_energy(self):
        K = 0.73*10**-10*(math.sqrt((2*NUM_AVOG*ELECTRON_CHARGE**2*I)/(Esolv*Evac*R*T)))
        alpha = math.sqrt(self.sgbnpType1*self.sgbnpType2)
        D = (self.radnpSGB1**2/(2*alpha)**2)
        fGB = math.sqrt(self.radnpSGB1**2 + alpha**2 * math.exp(-D))
        energy = ((1/Ein) - (math.exp(-K*fGB) / Esolv)) * ((self.charge1 * self.charge2) / fGB) \
                 - ((1/Ein) - (math.exp(-K*self.sgbnpType1)/Esolv))*(self.charge1**2/self.sgbnpType1)
        return energy


class PlotEnergies:
    def __init__(self, template, atom1, atom2):
        self.template = template
        self.atom1 = atom1
        self.atom2 = atom2
        self.nbon_energy = ComputeNBONEnergies(epsilon1=self.template.list_of_atoms[self.atom1].epsilon,
                                               epsilon2=self.template.list_of_atoms[self.atom2].epsilon,
                                               sigma1=self.template.list_of_atoms[self.atom1].sigma,
                                               sigma2=self.template.list_of_atoms[self.atom2].sigma,
                                               charge1=self.template.list_of_atoms[self.atom2].charge,
                                               charge2=self.template.list_of_atoms[self.atom2].charge,
                                               radnpSGB1=self.template.list_of_atoms[self.atom1].radnpSGB,
                                               radnpSGB2=self.template.list_of_atoms[self.atom2].radnpSGB,
                                               sgbnpType1=self.template.list_of_atoms[self.atom1].sgbnpType,
                                               sgbnpType2=self.template.list_of_atoms[self.atom2].sgbnpType)

    def plot_vdw_function(self):
        t1 = np.arange(0.01, 5.0, 0.001)
        energies = []
        for radii in t1:
            self.nbon_energy.radii = radii
            energy = self.nbon_energy.energy_vdw()
            energies.append(energy)
        return plt.plot(t1, energies, lw=2)

    def plot_charge_function(self):
        t1 = np.arange(0.01, 5.0, 0.001)
        energies = []
        for radii in t1:
            self.nbon_energy.radii = radii
            energy = self.nbon_energy.charge_function()
            energies.append(energy)
        return plt.plot(t1, energies, lw=2)

    def plot_solv_function(self):
        t1 = np.arange(0.01, 5.0, 0.001)
        energies = []
        for radii in t1:
            self.nbon_energy.radii = radii
            energy = self.nbon_energy.solv_pol_energy()
            energies.append(energy)
        return plt.plot(t1, energies, lw=2)

    def plot_nbon_function(self):
        t1 = np.arange(0.01, 5.0, 0.001)
        energies = []
        for radii in t1:
            self.nbon_energy.radii = radii
            energy = self.nbon_energy.energy_vdw()
            energy += self.nbon_energy.charge_function()
            energy += self.nbon_energy.solv_pol_energy()
            energies.append(energy)
        return plt.plot(t1, energies, lw=2)

    def show_plots(self):
        plt.figure(1)
        plt.subplot(221)
        self.plot_vdw_function()
        plt.subplot(222)
        self.plot_charge_function()
        plt.subplot(223)
        self.plot_solv_function()
        plt.subplot(224)
        self.plot_nbon_function()
        plt.show()



import sys
import re
import os
import logging
# Local imports 
import frag_pele.constants as c
import frag_pele.Helpers.checker as ch
# Getting the name of the module for the log system
logger = logging.getLogger(__name__)


def read_instructions_from_file(file):
    """
    It reads an "instruction file". This file contains the information of all the growing's that the user wants to
    perform, each one separated by newlines. Each instruction must have at least 3 columns, separated by tabulations:
    (1) name of the fragment's PDB file,
    (2) atom name of the core (if the user wants to add the H it must be separated by "-", example: C4-H2),
    (3) atom name of the fragment.
    If the line has more than 3 columns, it will be considered as successive growing's simulations, following the same
    structure previously mentioned but several times.
    :param file: filename of the instructions file.
    :return: list with all instructions processed.
    """
    list_of_instructions = []
    with open(file) as sf:
        instructions = sf.readlines()
        for line in instructions:
            if 0 < len(line.split()) <= 3:
                # Read from file the required information
                try:
                    fragment_pdb = line.split()[0]
                    core_atom = line.split()[1]
                    fragment_atom = line.split()[2]
                    ID = "{}{}{}".format(os.path.splitext(fragment_pdb)[0], core_atom, fragment_atom)
                    task = (fragment_pdb, core_atom, fragment_atom, ID)
                    list_of_instructions.append(task)
                except IndexError:
                    logger.critical("Check that the serie file {} contains: 'PDB_fragment_file'\t'PDB_core_atom_name'\t'PDB_fragment_atom_name' ".format(file))
            elif len(line.split()) > 3:
                growing_counter = len(line.split()) / 3
                successive_tasks = []
                for i in range(int(growing_counter)):
                    try:
                        fragment_pdb = line.split()[i * 3]
                        core_atom = line.split()[(i * 3) + 1]
                        if "*" in core_atom:
                            # Get fragment number and remove part of string
                            fragment_number = re.findall(r'[*]\d[*]', core_atom)[0].strip("*")
                            core_atom = core_atom.replace("*{}*".format(fragment_number), "")
                        else:
                            fragment_number = None
                        fragment_atom = line.split()[(i * 3) + 2]
                        ID = "{}{}{}".format(os.path.splitext(fragment_pdb)[0], core_atom, fragment_atom)
                        task = (fragment_pdb, core_atom, fragment_atom, ID, fragment_number)
                        successive_tasks.append(task)
                    except IndexError:
                        logger.critical("Check that the serie file {} contains: 'PDB_fragment_file'\t'PDB_core_atom_name'\t'PDB_fragment_atom_name' ".format(file))
                list_of_instructions.append(successive_tasks)
    return list_of_instructions


def get_pdb_fragments_and_atoms_from_instructions(list_of_instructions):
    """
    Given a list with the instructions processed it returns a list with containing the following elements:
    [fragment_pdb_file, pdb_atom_name_of_the_core (and H if required), pdb_atom_name_of_the_fragment (and H if required)]
    :param list_of_instructions: list with the instructions read from the instructions file. list
    :return: list with all fragments and atoms that will be used in the future. list
    """
    fragments_pdbs_and_atoms = []
    for instructions in list_of_instructions:
        if type(instructions) == list:
            for single_instruction in instructions:
                fragment_pdb, atom_core, atom_fragment = single_instruction[0], single_instruction[1], single_instruction[2]
                fragments_pdbs_and_atoms.append((fragment_pdb, atom_core, atom_fragment))
        else:
            fragment_pdb, atom_core, atom_fragment = instructions[0], instructions[1], instructions[2]
            fragments_pdbs_and_atoms.append((fragment_pdb, atom_core, atom_fragment))

    return fragments_pdbs_and_atoms

    
def check_instructions(instructions, complex_pdb, c_chain = "L", f_chain="L"):
    """
    It checks if the selected atoms exists in their correspondent PDB file and also checks if there are repeated
    PDB-atom-names in the PDB file.
    :param list_of_instructions: list with the instructions read from the instructions file. list
    :param complex: PDB file with the complex that contains the core
    :return: if something is wrong it raises an exception.
    """
    fragments_and_atoms = get_pdb_fragments_and_atoms_from_instructions([instructions])
    for fragment, atom_core, atom_fr in fragments_and_atoms:
        atoms_if_bond = extract_hydrogens_from_instructions([fragment, atom_core, atom_fr])
        if atoms_if_bond:
            ch.check_if_atom_exists_in_ligand(complex_pdb, atoms_if_bond[0], c_chain)
            ch.check_if_atom_exists_in_ligand(complex_pdb, atoms_if_bond[1], c_chain)
            ch.check_if_atom_exists_in_ligand(fragment, atoms_if_bond[2], f_chain)
            ch.check_if_atom_exists_in_ligand(fragment, atoms_if_bond[3], f_chain)
        else:
            ch.check_if_atom_exists_in_ligand(fragment, atom_fr, f_chain)
            ch.check_if_atom_exists_in_ligand(complex_pdb, atom_core, c_chain)
        with open(fragment) as content:
            fr_content = content.readlines()
        ch.check_duplicated_pdbatomnames(fr_content)


def extract_hydrogens_from_instructions(instruction):
    """
    If the core or the fragment atom contains a "-" means that the user is selecting an specific H to be bonded with
    the heavy atom. For this reason, this detects if this option has been selected by the user and extract the PDB-atom-name
    of each element.
    :param instruction: list that follow this structure: [fragment_pdb_file, pdb_atom_name_of_the_core
    (and H if required), pdb_atom_name_of_the_fragment (and H if required)]
    :return: if the "-" is found in the pdb_atom_name_of_the_core or pdb_atom_name_of_the_fragment it returns a list with
    PDB-atom-names split following this order: heavy_atom_core, hydrogen_core, heavy_atom_fragment, hydrogen_fragment.
    Otherwise it returns False.
    """
    if "-" in instruction[1] or "-" in instruction[2]:
        try:
            heavy_core = instruction[1].split("-")[0]
            hydrogen_core = instruction[1].split("-")[1]
            heavy_fragment = instruction[2].split("-")[0]
            hydrogen_fragment = instruction[2].split("-")[1]
            return heavy_core, hydrogen_core, heavy_fragment, hydrogen_fragment
        except IndexError:
            raise IndexError("To use steriochemistry the hydrogen from the core AND fragmet must be specified via controlfile")
    else:
        return False




import argparse
import numpy as np
import os
from AdaptivePELE.utilities import utilities


def parseArgs():
    parser = argparse.ArgumentParser(description="Get representative structures "
                                     "by selecting the ones closest to the "
                                     "cluster centers")
    parser.add_argument('representatives', type=str, help="Path to the representative structures summary file")
    parser.add_argument('structures_path', type=str, help="Path to the simulation structures")
    parser.add_argument('-o', type=str, default="", help="Path where to store the structures")
    parser.add_argument('-c', '--clusters', default=['a'], nargs="*", help="Clusters to extract, if not specified it will extract all")
    parser.add_argument("-t", "-trajectoryName", type=str, default="trajectory", help="Name of the trajectory files, e.g for trajectory_1.pdb the name is trajectory")
    args = parser.parse_args()
    return args


def main(representatives_files, path_structures, output="", clusters=None, trajNames="trajectory"):
    if clusters is None:
        clusters = ['a']
    # Load the representative structures file
    try:
        clusters_info = np.loadtxt(representatives_files, skiprows=1)
    except IOError:
        raise IOError("Couldn't find a representative file in %s, please check that the path is correct" % representatives_files)
    # Organize to minimise pdb loading
    if clusters != ['a']:
        clusters_info = clusters_info[map(int, clusters)]

    extract_info = {}
    for row in clusters_info:
        index = (row[1], row[2])
        value = (row[0], row[3])
        if index in extract_info:
            extract_info[index].append(value)
        else:
            extract_info[index] = [value]

    # Write appropiate pdbs
    destFolder = output
    if not output:
        destFolder, _ = os.path.split(representatives_files)
        destFolder = os.path.join(destFolder, "representative_structures_pdbs")

    if not os.path.exists(destFolder):
        os.makedirs(destFolder)
    else:
        destFolder += "_%d"
        it = 1
        while os.path.exists(destFolder % it):
            it += 1
        destFolder %= it
        os.makedirs(destFolder)
    structureFolder = os.path.join(path_structures, "%d", trajNames+"_%d.pdb")
    for trajFile, extraInfo in extract_info.items():
        pdbFile = structureFolder % trajFile
        try:
            snapshots = utilities.getSnapshots(pdbFile)
        except IOError:
            raise IOError("Unable to open %s, please check that the path to structures provided is correct" % pdbFile)
        for pair in extraInfo:
            with open(os.path.join(destFolder, "cluster_%d.pdb" % pair[0]), "w") as fw:
                fw.write(snapshots[int(pair[1])])
                fw.write("\n")


if __name__ == "__main__":
    arg = parseArgs()
    main(arg.representatives, arg.structures_path, arg.o, arg.clusters, arg.t)



"""
    Only for developers.
    Change releaseName to build a new release.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import socket
import glob
import shutil
import os
import subprocess


def copy_ignore(src, names):
    return [x for x in names if x.endswith(".c") or x.endswith(".so")]


machine = socket.gethostname()
if "bsccv" in machine:
    releaseFolder = "/data2/bsc72/AdaptiveSampling/bin"
elif 'login' in machine:
    name = os.getenv("BSC_MACHINE")
    if name == "mn4":
        releaseFolder = "/gpfs/projects/bsc72/adaptiveSampling/bin"
    elif name == "nord3":
        releaseFolder = "/gpfs/projects/bsc72/adaptiveSampling/bin_nord"

releaseName = "v1.4.2"
toOmit = ["tests", "runAllTests.py", "os", "sys", "TODO.txt", "Data", "Documents", "DataLocal", "epsilon_values.txt", "makeRelease.py", ".git", ".gitignore"]


files = glob.glob("*")

destFolder = os.path.join(releaseFolder, releaseName, "AdaptivePELE", "%s")
for filename in files:
    if filename in toOmit or filename.startswith(".") or filename.endswith("pyc"):
        continue
    try:
        if not os.path.exists(destFolder % filename):
            print("Copying", filename)
            shutil.copytree(filename, destFolder % filename, ignore=copy_ignore)
    except (IOError, OSError):
        shutil.copyfile(filename, destFolder % filename)

extraFiles = ["../README.rst", "../setup.py"]
for filename in extraFiles:
    if not os.path.exists(destFolder % filename):
        shutil.copyfile(filename, destFolder % filename)
        print("Copying", os.path.split(filename)[1])

print("Compiling cython extensions")
os.chdir(destFolder % "..")
subprocess.call(['python', 'setup.py', 'build_ext', '--inplace'])

print("Done with release %s!" % releaseName)
